# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasDashboardTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardQuery as RDSDashboardQuery

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import (
    add_entity_relationship, get_entity_attrs, get_entity_relationships,
)
from databuilder.utils.atlas import AtlasSerializedEntityOperation

LOGGER = logging.getLogger(__name__)


class DashboardQuery(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    A model that encapsulate Dashboard's query name
    """
    DASHBOARD_QUERY_LABEL = 'Query'
    DASHBOARD_QUERY_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group_id}/' \
                                 '{dashboard_id}/query/{query_id}'
    DASHBOARD_QUERY_RELATION_TYPE = 'HAS_QUERY'
    QUERY_DASHBOARD_RELATION_TYPE = 'QUERY_OF'

    def __init__(self,
                 dashboard_group_id: Optional[str],
                 dashboard_id: Optional[str],
                 query_name: str,
                 query_id: Optional[str] = None,
                 url: Optional[str] = '',
                 query_text: Optional[str] = None,
                 product: Optional[str] = '',
                 cluster: str = 'gold',
                 **kwargs: Any
                 ) -> None:
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        self._query_name = query_name
        self._query_id = query_id if query_id else query_name
        self._url = url
        self._query_text = query_text
        self._product = product
        self._cluster = cluster
        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()
        self._record_iterator = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes = {
            'id': self._query_id,
            'name': self._query_name,
        }

        if self._url:
            node_attributes['url'] = self._url

        if self._query_text:
            node_attributes['query_text'] = self._query_text

        node = GraphNode(
            key=self._get_query_node_key(),
            label=DashboardQuery.DASHBOARD_QUERY_LABEL,
            attributes=node_attributes
        )

        yield node

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relationship = GraphRelationship(
            start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            end_label=DashboardQuery.DASHBOARD_QUERY_LABEL,
            start_key=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            ),
            end_key=self._get_query_node_key(),
            type=DashboardQuery.DASHBOARD_QUERY_RELATION_TYPE,
            reverse_type=DashboardQuery.QUERY_DASHBOARD_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        record = RDSDashboardQuery(
            rk=self._get_query_node_key(),
            id=self._query_id,
            name=self._query_name,
            dashboard_rk=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            )
        )
        if self._url:
            record.url = self._url
        if self._query_text:
            record.query_text = self._query_text

        yield record

    def _get_query_node_key(self) -> str:
        return DashboardQuery.DASHBOARD_QUERY_KEY_FORMAT.format(
            product=self._product,
            cluster=self._cluster,
            dashboard_group_id=self._dashboard_group_id,
            dashboard_id=self._dashboard_id,
            query_id=self._query_id
        )

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        # Query
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_query_node_key()),
            ('name', self._query_name),
            ('id', self._query_id),
            ('url', self._url),
            ('queryText', self._query_text)
        ]

        query_entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'dashboard',
            AtlasDashboardTypes.metadata,
            DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            )
        )

        query_entity = AtlasEntity(
            typeName=AtlasDashboardTypes.query,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=query_entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )
        yield query_entity

    def __repr__(self) -> str:
        return f'DashboardQuery({self._dashboard_group_id!r}, {self._dashboard_id!r}, {self._query_name!r}, ' \
               f'{self._query_id!r}, {self._url!r}, {self._query_text!r}, {self._product!r}, {self._cluster!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator, Optional

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.query.query import QueryMetadata
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class QueryJoinMetadata(GraphSerializable):
    """
    A Join clause used between two tables within a query
    """
    NODE_LABEL = 'Join'
    KEY_FORMAT = '{join_type}-{left_column_key}-{operator}-{right_column_key}'

    # Relation between entity and query
    COLUMN_JOIN_RELATION_TYPE = 'COLUMN_JOINS_WITH'
    INVERSE_COLUMN_JOIN_RELATION_TYPE = 'JOIN_OF_COLUMN'

    QUERY_JOIN_RELATION_TYPE = 'QUERY_JOINS_WITH'
    INVERSE_QUERY_JOIN_RELATION_TYPE = 'JOIN_OF_QUERY'

    # Node attributes
    JOIN_TYPE = 'join_type'
    JOIN_OPERATOR = 'operator'
    JOIN_SQL = 'join_sql'
    LEFT_TABLE_KEY = 'left_table_key'
    LEFT_DATABASE = 'left_database'
    LEFT_CLUSTER = 'left_cluster'
    LEFT_SCHEMA = 'left_schema'
    LEFT_TABLE = 'left_table'
    RIGHT_TABLE_KEY = 'right_table_key'
    RIGHT_DATABASE = 'right_database'
    RIGHT_CLUSTER = 'right_cluster'
    RIGHT_SCHEMA = 'right_schema'
    RIGHT_TABLE = 'right_table'

    def __init__(self,
                 left_table: TableMetadata,
                 right_table: TableMetadata,
                 left_column: ColumnMetadata,
                 right_column: ColumnMetadata,
                 join_type: str,
                 join_operator: str,
                 join_sql: str,
                 query_metadata: Optional[QueryMetadata] = None,
                 yield_relation_nodes: bool = False):
        """
        :param left_table: The table joined on the left side of the join clause
        :param right_table: The table joined on the right side of the join clause
        :param left_column: The column from the left table used in the join
        :param right_column: The column from the right table used in the join
        :param join_type: A free form string representing the type of join, examples
            include: inner join, right join, full join, etc.
        :param join_operator: The operator used in the join, examples include: =, >, etc.
        :param query_metadata: The Query metadata object that this where clause belongs to, this
            is optional to allow creating static QueryJoinMetadata objects to show on tables
            without the complexity of creating QueryMetadata
        :param yield_relation_nodes: A boolean, indicating whether or not the query metadata
            and tables associated to this Join should have nodes created if they does not
            already exist.
        """
        # For inner joins we don't want to duplicate joins if the other table
        # comes first in the join clause since it produces the same effect.
        # This ONLY applies to inner join and you may need to massage your data
        # for join_type to have the proper value
        swap_left_right = False
        if join_operator == '=' and join_type == 'inner join':
            tables_sorted = sorted([left_table._get_table_key(), right_table._get_table_key()])
            if tables_sorted[0] == right_table:
                swap_left_right = True

        self.left_table = right_table if swap_left_right else left_table
        self.right_table = left_table if swap_left_right else right_table
        self.left_column = right_column if swap_left_right else left_column
        self.right_column = left_column if swap_left_right else right_column

        self.join_type = join_type
        self.join_operator = join_operator
        self.join_sql = join_sql
        self.query_metadata = query_metadata
        self.yield_relation_nodes = yield_relation_nodes
        self._node_iter = self._create_next_node()
        self._relation_iter = self._create_relation_iterator()

    def __repr__(self) -> str:
        return (
            f'QueryJoinMetadata(Left Table: {self.left_table._get_table_key()}, '
            f'Right Table: {self.left_table._get_table_key()})'
        )

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    @staticmethod
    def get_key(left_column_key: str, right_column_key: str, join_type: str, operator: str) -> str:
        join_no_space = join_type.replace(' ', '-')
        return QueryJoinMetadata.KEY_FORMAT.format(left_column_key=left_column_key,
                                                   right_column_key=right_column_key,
                                                   join_type=join_no_space,
                                                   operator=operator)

    def get_key_self(self) -> str:
        return QueryJoinMetadata.get_key(left_column_key=self.left_table._get_col_key(col=self.left_column),
                                         right_column_key=self.right_table._get_col_key(col=self.right_column),
                                         join_type=self.join_type,
                                         operator=self.join_operator)

    def get_query_relations(self) -> Iterator[GraphRelationship]:

        # Left Column
        yield GraphRelationship(
            start_label=ColumnMetadata.COLUMN_NODE_LABEL,
            end_label=self.NODE_LABEL,
            start_key=self.left_table._get_col_key(col=self.left_column),
            end_key=self.get_key_self(),
            type=self.COLUMN_JOIN_RELATION_TYPE,
            reverse_type=self.INVERSE_COLUMN_JOIN_RELATION_TYPE,
            attributes={}
        )

        # Right Column
        yield GraphRelationship(
            start_label=ColumnMetadata.COLUMN_NODE_LABEL,
            end_label=self.NODE_LABEL,
            start_key=self.right_table._get_col_key(col=self.right_column),
            end_key=self.get_key_self(),
            type=self.COLUMN_JOIN_RELATION_TYPE,
            reverse_type=self.INVERSE_COLUMN_JOIN_RELATION_TYPE,
            attributes={}
        )

        if self.query_metadata:
            yield GraphRelationship(
                start_label=QueryMetadata.NODE_LABEL,
                end_label=self.NODE_LABEL,
                start_key=self.query_metadata.get_key_self(),
                end_key=self.get_key_self(),
                type=self.QUERY_JOIN_RELATION_TYPE,
                reverse_type=self.INVERSE_QUERY_JOIN_RELATION_TYPE,
                attributes={}
            )

    def _create_next_node(self) -> Iterator[GraphNode]:
        """
        Create query nodes
        :return:
        """
        yield GraphNode(
            key=self.get_key_self(),
            label=self.NODE_LABEL,
            attributes={
                self.JOIN_TYPE: self.join_type,
                self.JOIN_OPERATOR: self.join_operator,
                self.JOIN_SQL: self.join_sql,
                self.LEFT_TABLE_KEY: self.left_table._get_table_key(),
                self.LEFT_DATABASE: self.left_table.database,
                self.LEFT_CLUSTER: self.left_table.cluster,
                self.LEFT_SCHEMA: self.left_table.schema,
                self.LEFT_TABLE: self.left_table.name,
                self.RIGHT_TABLE_KEY: self.right_table._get_table_key(),
                self.RIGHT_DATABASE: self.right_table.database,
                self.RIGHT_CLUSTER: self.right_table.cluster,
                self.RIGHT_SCHEMA: self.right_table.schema,
                self.RIGHT_TABLE: self.right_table.name
            }
        )

        if self.yield_relation_nodes:
            for l_tbl_item in self.left_table._create_next_node():
                yield l_tbl_item
            for r_tbl_item in self.right_table._create_next_node():
                yield r_tbl_item
            if self.query_metadata:
                for query_item in self.query_metadata._create_next_node():
                    yield query_item

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relations = self.get_query_relations()
        for relation in relations:
            yield relation

        if self.yield_relation_nodes:
            for l_tbl_rel in self.left_table._create_next_relation():
                yield l_tbl_rel
            for r_tbl_rel in self.right_table._create_next_relation():
                yield r_tbl_rel
            if self.query_metadata:
                for query_rel in self.query_metadata._create_relation_iterator():
                    yield query_rel



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

AtlasEntity = namedtuple(
    'AtlasEntity',
    [
        'operation',
        'typeName',
        'relationships',
        'attributes'
    ]
)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import time
from datetime import datetime
from functools import wraps
from multiprocessing.pool import ThreadPool
from typing import (
    Any, Iterator, List, Union,
)

from pyhocon import ConfigFactory, ConfigTree
from pytz import UTC

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.filesystem.filesystem import FileSystem, is_client_side_error
from databuilder.models.table_last_updated import TableLastUpdated

LOGGER = logging.getLogger(__name__)
OLDEST_TIMESTAMP = datetime.fromtimestamp(0, UTC)


def fs_error_handler(f: Any) -> Any:
    """
    A Decorator that handles error from FileSystem for HiveTableLastUpdatedExtractor use case
    If it's client side error, it logs in INFO level, and other errors is logged as error level with stacktrace.
    The decorator is intentionally not re-raising exception so that it can isolate the error.
    :param f:
    :return:
    """

    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except Exception as e:
            if is_client_side_error(e):
                LOGGER.info('Invalid metadata. Skipping. args: %s, kwargs: %s. error: %s', args, kwargs, e)
                return None
            else:
                LOGGER.exception('Unknown exception while processing args: %s, kwargs: %s', args, kwargs)
                return None

    return wrapper


class HiveTableLastUpdatedExtractor(Extractor):
    """
    Uses Hive metastore and underlying storage to figure out last updated timestamp of table.

    It turned out that we cannot use table param "last_modified_time", as it provides DDL date not DML date.
    For this reason, we are utilizing underlying file of Hive to fetch latest updated date.
    However, it is not efficient to poke all files in Hive, and it only poke underlying storage for non-partitioned
    table. For partitioned table, it will fetch partition created timestamp, and it's close enough for last updated
    timestamp.

    """
    DEFAULT_PARTITION_TABLE_SQL_STATEMENT = """
    SELECT
    DBS.NAME as `schema`,
    TBL_NAME as table_name,
    MAX(PARTITIONS.CREATE_TIME) as last_updated_time
    FROM TBLS
    JOIN DBS ON TBLS.DB_ID = DBS.DB_ID
    JOIN PARTITIONS ON TBLS.TBL_ID = PARTITIONS.TBL_ID
    {where_clause_suffix}
    GROUP BY `schema`, table_name
    ORDER BY `schema`, table_name;
    """

    DEFAULT_POSTGRES_PARTITION_TABLE_SQL_STATEMENT = """
    SELECT
    d."NAME" as "schema",
    t."TBL_NAME" as table_name,
    MAX(p."CREATE_TIME") as last_updated_time
    FROM "TBLS" t
    JOIN "DBS" d ON t."DB_ID" = d."DB_ID"
    JOIN "PARTITIONS" p ON t."TBL_ID" = p."TBL_ID"
    {where_clause_suffix}
    GROUP BY "schema", table_name
    ORDER BY "schema", table_name;
    """

    DEFAULT_NON_PARTITIONED_TABLE_SQL_STATEMENT = """
    SELECT
    DBS.NAME as `schema`,
    TBL_NAME as table_name,
    SDS.LOCATION as location
    FROM TBLS
    JOIN DBS ON TBLS.DB_ID = DBS.DB_ID
    JOIN SDS ON TBLS.SD_ID = SDS.SD_ID
    {where_clause_suffix}
    ORDER BY `schema`, table_name;
    """

    DEFAULT_POSTGRES_NON_PARTITIONED_TABLE_SQL_STATEMENT = """
    SELECT
    d."NAME" as "schema",
    t."TBL_NAME" as table_name,
    s."LOCATION" as location
    FROM "TBLS" t
    JOIN "DBS" d ON t."DB_ID" = d."DB_ID"
    JOIN "SDS" s ON t."SD_ID" = s."SD_ID"
    {where_clause_suffix}
    ORDER BY "schema", table_name;
    """

    # Additional where clause for non partitioned table SQL
    DEFAULT_ADDTIONAL_WHERE_CLAUSE = """ NOT EXISTS (SELECT * FROM PARTITIONS WHERE PARTITIONS.TBL_ID = TBLS.TBL_ID)
    AND NOT EXISTS (SELECT * FROM PARTITION_KEYS WHERE PARTITION_KEYS.TBL_ID = TBLS.TBL_ID)
    """

    DEFAULT_POSTGRES_ADDTIONAL_WHERE_CLAUSE = """ NOT EXISTS (SELECT * FROM "PARTITIONS" p
    WHERE p."TBL_ID" = t."TBL_ID") AND NOT EXISTS (SELECT * FROM "PARTITION_KEYS" pk WHERE pk."TBL_ID" = t."TBL_ID")
    """

    DATABASE = 'hive'

    # CONFIG KEYS
    PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY = 'partitioned_table_where_clause_suffix'
    NON_PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY = 'non_partitioned_table_where_clause_suffix'
    CLUSTER_KEY = 'cluster'
    # number of threads that fetches metadata from FileSystem
    FS_WORKER_POOL_SIZE = 'fs_worker_pool_size'
    FS_WORKER_TIMEOUT_SEC = 'fs_worker_timeout_sec'
    # If number of files that it needs to fetch metadata is larger than this threshold, it will skip the table.
    FILE_CHECK_THRESHOLD = 'file_check_threshold'

    DEFAULT_CONFIG = ConfigFactory.from_dict({PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY: ' ',
                                              NON_PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY: ' ',
                                              CLUSTER_KEY: 'gold',
                                              FS_WORKER_POOL_SIZE: 500,
                                              FS_WORKER_TIMEOUT_SEC: 60,
                                              FILE_CHECK_THRESHOLD: -1})

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf.with_fallback(HiveTableLastUpdatedExtractor.DEFAULT_CONFIG)

        pool_size = self._conf.get_int(HiveTableLastUpdatedExtractor.FS_WORKER_POOL_SIZE)
        LOGGER.info('Using thread pool size: %s', pool_size)
        self._fs_worker_pool = ThreadPool(processes=pool_size)
        self._fs_worker_timeout = self._conf.get_int(HiveTableLastUpdatedExtractor.FS_WORKER_TIMEOUT_SEC)
        LOGGER.info('Using thread timeout: %s seconds', self._fs_worker_timeout)

        self._cluster = self._conf.get_string(HiveTableLastUpdatedExtractor.CLUSTER_KEY)

        self._partitioned_table_extractor = self._get_partitioned_table_sql_alchemy_extractor()
        self._non_partitioned_table_extractor = self._get_non_partitioned_table_sql_alchemy_extractor()
        self._fs = self._get_filesystem()
        self._last_updated_filecheck_threshold \
            = self._conf.get_int(HiveTableLastUpdatedExtractor.FILE_CHECK_THRESHOLD)

        self._extract_iter: Union[None, Iterator] = None

    def _get_partitioned_table_sql_alchemy_extractor(self) -> Extractor:
        """
        Getting an SQLAlchemy extractor that extracts last updated timestamp for partitioned table.
        :return: SQLAlchemyExtractor
        """

        sql_stmt = self._choose_default_partitioned_sql_stm().format(
            where_clause_suffix=self._conf.get_string(
                self.PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY, ' '))

        LOGGER.info('SQL for partitioned table against Hive metastore: %s', sql_stmt)

        sql_alchemy_extractor = SQLAlchemyExtractor()
        sql_alchemy_conf = Scoped.get_scoped_conf(self._conf, sql_alchemy_extractor.get_scope()) \
            .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: sql_stmt}))
        sql_alchemy_extractor.init(sql_alchemy_conf)
        return sql_alchemy_extractor

    def _choose_default_partitioned_sql_stm(self) -> str:
        conn_string = self._conf.get_string("extractor.sqlalchemy.conn_string")
        if conn_string.startswith('postgres') or conn_string.startswith('postgresql'):
            return self.DEFAULT_POSTGRES_PARTITION_TABLE_SQL_STATEMENT
        else:
            return self.DEFAULT_PARTITION_TABLE_SQL_STATEMENT

    def _get_non_partitioned_table_sql_alchemy_extractor(self) -> Extractor:
        """
        Getting an SQLAlchemy extractor that extracts storage location for non-partitioned table for further probing
        last updated timestamp

        :return: SQLAlchemyExtractor
        """

        default_sql_stmt, default_additional_where_clause = self._choose_default_non_partitioned_sql_stm()

        if self.NON_PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY in self._conf:
            where_clause_suffix = """
            {}
            AND {}
            """.format(self._conf.get_string(
                self.NON_PARTITIONED_TABLE_WHERE_CLAUSE_SUFFIX_KEY),
                default_additional_where_clause)
        else:
            where_clause_suffix = 'WHERE {}'.format(default_additional_where_clause)

        sql_stmt = default_sql_stmt.format(
            where_clause_suffix=where_clause_suffix)

        LOGGER.info('SQL for non-partitioned table against Hive metastore: %s', sql_stmt)

        sql_alchemy_extractor = SQLAlchemyExtractor()
        sql_alchemy_conf = Scoped.get_scoped_conf(self._conf, sql_alchemy_extractor.get_scope()) \
            .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: sql_stmt}))
        sql_alchemy_extractor.init(sql_alchemy_conf)
        return sql_alchemy_extractor

    def _choose_default_non_partitioned_sql_stm(self) -> List[str]:
        conn_string = self._conf.get_string("extractor.sqlalchemy.conn_string")
        if conn_string.startswith('postgres') or conn_string.startswith('postgresql'):
            return [self.DEFAULT_POSTGRES_NON_PARTITIONED_TABLE_SQL_STATEMENT,
                    self.DEFAULT_POSTGRES_ADDTIONAL_WHERE_CLAUSE]
        else:
            return [self.DEFAULT_NON_PARTITIONED_TABLE_SQL_STATEMENT, self.DEFAULT_ADDTIONAL_WHERE_CLAUSE]

    def _get_filesystem(self) -> FileSystem:
        fs = FileSystem()
        fs.init(Scoped.get_scoped_conf(self._conf, fs.get_scope()))
        return fs

    def extract(self) -> Union[TableLastUpdated, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.hive_table_last_updated'

    def _get_extract_iter(self) -> Iterator[TableLastUpdated]:
        """
        An iterator that utilizes Generator pattern. First it provides TableLastUpdated objects for partitioned
        table, straight from partitioned_table_extractor (SQLAlchemyExtractor)

        Once partitioned table is done, it uses non_partitioned_table_extractor to get storage location of table,
        and probing files under storage location to get max timestamp per table.
        :return:
        """

        partitioned_tbl_row = self._partitioned_table_extractor.extract()
        while partitioned_tbl_row:
            yield TableLastUpdated(table_name=partitioned_tbl_row['table_name'],
                                   last_updated_time_epoch=partitioned_tbl_row['last_updated_time'],
                                   schema=partitioned_tbl_row['schema'],
                                   db=HiveTableLastUpdatedExtractor.DATABASE,
                                   cluster=self._cluster)
            partitioned_tbl_row = self._partitioned_table_extractor.extract()

        LOGGER.info('Extracting non-partitioned table')
        count = 0
        non_partitioned_tbl_row = self._non_partitioned_table_extractor.extract()
        while non_partitioned_tbl_row:
            count += 1
            if count % 10 == 0:
                LOGGER.info('Processed %i non-partitioned tables', count)

            if not non_partitioned_tbl_row['location']:
                LOGGER.warning('Skipping as no storage location available. %s', non_partitioned_tbl_row)
                non_partitioned_tbl_row = self._non_partitioned_table_extractor.extract()
                continue

            start = time.time()
            table_last_updated = self._get_last_updated_datetime_from_filesystem(
                table=non_partitioned_tbl_row['table_name'],
                schema=non_partitioned_tbl_row['schema'],
                storage_location=non_partitioned_tbl_row['location'])
            LOGGER.info('Elapsed: %i seconds', time.time() - start)

            if table_last_updated:
                yield table_last_updated

            non_partitioned_tbl_row = self._non_partitioned_table_extractor.extract()

    def _get_last_updated_datetime_from_filesystem(self,
                                                   table: str,
                                                   schema: str,
                                                   storage_location: str,
                                                   ) -> Union[TableLastUpdated, None]:
        """
        Fetching metadata within files under storage_location to get latest timestamp.
        (First level only under storage_location)
        Utilizes thread pool to enhance performance. Not using processpool, as it's almost entirely IO bound operation.

        :param table:
        :param schema:
        :param storage_location:
        :return:
        """

        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(f'Getting last updated datetime for {schema}.{table} in {storage_location}')

        last_updated = OLDEST_TIMESTAMP

        paths = self._ls(storage_location)
        if not paths:
            LOGGER.info(f'{schema}.{table} does not have any file in path {storage_location}. Skipping')
            return None

        LOGGER.info(f'Fetching metadata for {schema}.{table} of {len(paths)} files')

        if 0 < self._last_updated_filecheck_threshold < len(paths):
            LOGGER.info(f'Skipping {schema}.{table} due to too many files. '
                        f'{len(paths)} files exist in {storage_location}')
            return None

        time_stamp_futures = \
            [self._fs_worker_pool.apply_async(self._get_timestamp, (path, schema, table, storage_location))
             for path in paths]
        for time_stamp_future in time_stamp_futures:
            try:
                time_stamp = time_stamp_future.get(timeout=self._fs_worker_timeout)
                if time_stamp:
                    last_updated = max(time_stamp, last_updated)
            except TimeoutError:
                LOGGER.warning('Timed out on paths %s . Skipping', paths)

        if last_updated == OLDEST_TIMESTAMP:
            LOGGER.info(f'No timestamp was derived on {schema}.{table} from location: {storage_location} . Skipping')
            return None

        result = TableLastUpdated(table_name=table,
                                  last_updated_time_epoch=int((last_updated - OLDEST_TIMESTAMP).total_seconds()),
                                  schema=schema,
                                  db=HiveTableLastUpdatedExtractor.DATABASE,
                                  cluster=self._cluster)

        return result

    @fs_error_handler
    def _ls(self, path: str) -> List[str]:
        """
        An wrapper to FileSystem.ls to use fs_error_handler decorator
        :param path:
        :return:
        """
        return self._fs.ls(path)

    @fs_error_handler
    def _get_timestamp(self,
                       path: str,
                       schema: str,
                       table: str,
                       storage_location: str,
                       ) -> Union[datetime, None]:
        """
        An wrapper to FileSystem.ls to use fs_error_handler decorator
        :param path:
        :param schema:
        :param table:
        :param storage_location:
        :return:
        """
        if not path:
            LOGGER.info(f'Empty path {path} on {schema}.{table} in storage location {storage_location} . Skipping')
            return None

        if not self._fs.is_file(path):
            return None

        file_metadata = self._fs.info(path)
        return file_metadata.last_updated



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import re
from typing import (
    Any, Iterator, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasTableTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.schema import (
    Schema as RDSSchema, SchemaDescription as RDSSchemaDescription,
    SchemaProgrammaticDescription as RDSSchemaProgrammaticDescription,
)

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.description_metadata import DescriptionMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.schema.schema_constant import (
    SCHEMA_KEY_PATTERN_REGEX, SCHEMA_NAME_ATTR, SCHEMA_NODE_LABEL,
)
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasSerializedEntityOperation


class SchemaModel(GraphSerializable, TableSerializable, AtlasSerializable):
    def __init__(self,
                 schema_key: str,
                 schema: str,
                 description: str = None,
                 description_source: str = None,
                 **kwargs: Any
                 ) -> None:
        self._schema_key = schema_key
        self._schema = schema
        self._cluster_key = self._get_cluster_key(schema_key)
        self._description = DescriptionMetadata.create_description_metadata(text=description,
                                                                            source=description_source) \
            if description else None
        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()
        self._record_iterator = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        node = GraphNode(
            key=self._schema_key,
            label=SCHEMA_NODE_LABEL,
            attributes={
                SCHEMA_NAME_ATTR: self._schema,
            }
        )
        yield node

        if self._description:
            yield self._description.get_node(self._get_description_node_key())

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        schema_record = RDSSchema(
            rk=self._schema_key,
            name=self._schema,
            cluster_rk=self._cluster_key
        )
        yield schema_record

        if self._description:
            if self._description.label == DescriptionMetadata.DESCRIPTION_NODE_LABEL:
                yield RDSSchemaDescription(
                    rk=self._get_description_node_key(),
                    description_source=self._description.source,
                    description=self._description.text,
                    schema_rk=self._schema_key
                )
            else:
                yield RDSSchemaProgrammaticDescription(
                    rk=self._get_description_node_key(),
                    description_source=self._description.source,
                    description=self._description.text,
                    schema_rk=self._schema_key
                )

    def _get_description_node_key(self) -> str:
        desc = self._description.get_description_id() if self._description is not None else ''
        return f'{self._schema_key}/{desc}'

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        if self._description:
            yield self._description.get_relation(start_node=SCHEMA_NODE_LABEL,
                                                 start_key=self._schema_key,
                                                 end_key=self._get_description_node_key())

    def _get_cluster_key(self, schema_key: str) -> str:
        schema_key_pattern = re.compile(SCHEMA_KEY_PATTERN_REGEX)
        schema_key_match = schema_key_pattern.match(schema_key)
        if not schema_key_match:
            raise Exception(f'{schema_key} does not match the schema key pattern')

        cluster_key = schema_key_match.group(1)
        return cluster_key

    def _create_atlas_schema_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._schema_key),
            ('name', self._schema_key),
            ('description', self._description.text if self._description else '')
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        # Since Schema cannot exist without Cluster (COMPOSITION relationship type), we assume Schema entity was created
        # by different process and we only update schema description here using UPDATE operation.
        entity = AtlasEntity(
            typeName=AtlasTableTypes.schema,
            operation=AtlasSerializedEntityOperation.UPDATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        yield self._create_atlas_schema_entity()

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Dict, Iterator,
)

from pyhocon import ConfigFactory, ConfigTree

import databuilder.extractor.dashboard.tableau.tableau_dashboard_constants as const
from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.tableau.tableau_dashboard_utils import (
    TableauDashboardUtils, TableauGraphQLApiExtractor,
)
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel

LOGGER = logging.getLogger(__name__)


class TableauGraphQLExternalTableExtractor(TableauGraphQLApiExtractor):
    """
    Implements the extraction-time logic for parsing the GraphQL result and transforming into a dict
    that fills the TableMetadata model.
    """

    EXTERNAL_CLUSTER_NAME = const.EXTERNAL_CLUSTER_NAME
    EXTERNAL_SCHEMA_NAME = const.EXTERNAL_SCHEMA_NAME

    def execute(self) -> Iterator[Dict[str, Any]]:
        response = self.execute_query()

        for table in response['databases']:
            if table['connectionType'] in ['google-sheets', 'salesforce', 'excel-direct']:
                for downstreamTable in table['tables']:
                    data = {
                        'cluster': self._conf.get_string(TableauGraphQLExternalTableExtractor.EXTERNAL_CLUSTER_NAME),
                        'database': TableauDashboardUtils.sanitize_database_name(
                            table['connectionType']
                        ),
                        'schema': TableauDashboardUtils.sanitize_schema_name(table['name']),
                        'name': TableauDashboardUtils.sanitize_table_name(downstreamTable['name']),
                        'description': table['description']
                    }
                    yield data
            else:
                data = {
                    'cluster': self._conf.get_string(TableauGraphQLExternalTableExtractor.EXTERNAL_CLUSTER_NAME),
                    'database': TableauDashboardUtils.sanitize_database_name(table['connectionType']),
                    'schema': self._conf.get_string(TableauGraphQLExternalTableExtractor.EXTERNAL_SCHEMA_NAME),
                    'name': TableauDashboardUtils.sanitize_table_name(table['name']),
                    'description': table['description']
                }
                yield data


class TableauDashboardExternalTableExtractor(Extractor):
    """
    Creates the "external" Tableau tables.
    In this context, "external" tables are "tables" that are not from a typical database, and are loaded
    using some other data format, like CSV files.
    This extractor has been tested with the following types of external tables:
        Excel spreadsheets
        Text files (including CSV files)
        Salesforce connections
        Google Sheets connections

    Excel spreadsheets, Salesforce connections, and Google Sheets connections are all classified as
    "databases" in terms of Tableau's Metadata API, with their "subsheets" forming their "tables" when
    present. However, these tables are not assigned a schema, this extractor chooses to use the name
    parent sheet as the schema, and assign a new table to each subsheet. The connection type is
    always used as the database, and for text files, the schema is set using the EXTERNAL_SCHEMA_NAME
    config option. Since these external tables are usually named for human consumption only and often
    contain a wider range of characters, all inputs are transformed to remove any problematic
    occurences before they are inserted: see the sanitize methods TableauDashboardUtils for specifics.

    A more concrete example: if one had a Google Sheet titled "Growth by Region & County" with 2 subsheets called
    "FY19 Report" and "FY20 Report", two tables would be generated with the following keys:
    googlesheets://external.growth_by_region_county/FY_19_Report
    googlesheets://external.growth_by_region_county/FY_20_Report
    """

    API_BASE_URL = const.API_BASE_URL
    API_VERSION = const.API_VERSION
    CLUSTER = const.CLUSTER
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS
    EXTERNAL_CLUSTER_NAME = const.EXTERNAL_CLUSTER_NAME
    EXTERNAL_SCHEMA_NAME = const.EXTERNAL_SCHEMA_NAME
    EXTERNAL_TABLE_TYPES = const.EXTERNAL_TABLE_TYPES
    SITE_NAME = const.SITE_NAME
    TABLEAU_ACCESS_TOKEN_NAME = const.TABLEAU_ACCESS_TOKEN_NAME
    TABLEAU_ACCESS_TOKEN_SECRET = const.TABLEAU_ACCESS_TOKEN_SECRET
    VERIFY_REQUEST = const.VERIFY_REQUEST

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf
        self.query = """query externalTables($externalTableTypes: [String]) {
          databases (filter: {connectionTypeWithin: $externalTableTypes}) {
            name
            connectionType
            description
            tables {
                name
            }
          }
        }"""
        self.query_variables = {
            'externalTableTypes': self._conf.get_list(TableauDashboardExternalTableExtractor.EXTERNAL_TABLE_TYPES)}
        self._extractor = self._build_extractor()

        transformers = []
        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.table_metadata.TableMetadata'})))
        transformers.append(dict_to_model_transformer)
        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self) -> Any:
        record = self._extractor.extract()
        if not record:
            return None

        return self._transformer.transform(record=record)

    def get_scope(self) -> str:
        return 'extractor.tableau_external_table'

    def _build_extractor(self) -> TableauGraphQLExternalTableExtractor:
        """
        Builds a TableauGraphQLExternalTableExtractor. All data required can be retrieved with a single GraphQL call.
        :return: A TableauGraphQLExternalTableExtractor that creates external table metadata entities.
        """
        extractor = TableauGraphQLExternalTableExtractor()

        config_dict = {
            TableauGraphQLApiExtractor.QUERY_VARIABLES: self.query_variables,
            TableauGraphQLApiExtractor.QUERY: self.query}
        tableau_extractor_conf = \
            Scoped.get_scoped_conf(self._conf, extractor.get_scope())\
                  .with_fallback(self._conf)\
                  .with_fallback(ConfigFactory.from_dict(config_dict))
        extractor.init(conf=tableau_extractor_conf)
        return extractor



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import datetime
import logging
import textwrap
import time
from calendar import timegm
from collections import namedtuple
from typing import (
    Any, Dict, Iterator, List, Tuple, Union,
)

from pyhocon import ConfigTree

from databuilder.extractor.base_bigquery_extractor import BaseBigQueryExtractor, DatasetRef
from databuilder.models.watermark import Watermark

PartitionInfo = namedtuple('PartitionInfo', ['partition_id', 'epoch_created'])

LOGGER = logging.getLogger(__name__)


class BigQueryWatermarkExtractor(BaseBigQueryExtractor):

    def init(self, conf: ConfigTree) -> None:
        BaseBigQueryExtractor.init(self, conf)
        self.iter: Iterator[Watermark] = iter(self._iterate_over_tables())

    def get_scope(self) -> str:
        return 'extractor.bigquery_watermarks'

    def _retrieve_tables(self,
                         dataset: DatasetRef
                         ) -> Iterator[Watermark]:
        sharded_table_watermarks: Dict[str, Dict[str, Union[str, Any]]] = {}
        cutoff_time_in_epoch = timegm(time.strptime(self.cutoff_time, BigQueryWatermarkExtractor.DATE_TIME_FORMAT))

        for page in self._page_table_list_results(dataset):
            if 'tables' not in page:
                continue

            for table in page['tables']:
                tableRef = table['tableReference']
                table_id = tableRef['tableId']
                table_creation_time = float(table['creationTime']) / 1000
                # only extract watermark metadata for tables created before the cut-off time
                if table_creation_time < cutoff_time_in_epoch:
                    # BigQuery tables that have numeric suffix starts with a date are
                    # considered date range tables.
                    # ( e.g. ga_sessions_20190101, ga_sessions_20190102, etc. )
                    # We use these dates in the suffixes to determine high and low watermarks
                    if self._is_sharded_table(table_id):
                        suffix = self._get_sharded_table_suffix(table_id)
                        prefix = table_id[:-len(suffix)]
                        date = suffix[:BaseBigQueryExtractor.DATE_LENGTH]

                        if prefix in sharded_table_watermarks:
                            sharded_table_watermarks[prefix]['low'] = min(
                                sharded_table_watermarks[prefix]['low'], date)
                            sharded_table_watermarks[prefix]['high'] = max(
                                sharded_table_watermarks[prefix]['high'], date)
                        else:
                            sharded_table_watermarks[prefix] = {'high': date, 'low': date, 'table': table}
                    else:
                        partitions = self._get_partitions(table, tableRef)
                        if not partitions:
                            continue
                        low, high = self._get_partition_watermarks(table, tableRef, partitions)
                        yield low
                        yield high

            for prefix, td in sharded_table_watermarks.items():
                table = td['table']
                tableRef = table['tableReference']

                yield Watermark(
                    datetime.datetime.fromtimestamp(float(table['creationTime']) / 1000).strftime('%Y-%m-%d %H:%M:%S'),
                    'bigquery',
                    tableRef['datasetId'],
                    prefix,
                    f'__table__={td["low"]}',
                    part_type="low_watermark",
                    cluster=tableRef['projectId']
                )

                yield Watermark(
                    datetime.datetime.fromtimestamp(float(table['creationTime']) / 1000).strftime('%Y-%m-%d %H:%M:%S'),
                    'bigquery',
                    tableRef['datasetId'],
                    prefix,
                    f'__table__={td["high"]}',
                    part_type="high_watermark",
                    cluster=tableRef['projectId']
                )

    def _get_partitions(self,
                        table: str,
                        tableRef: Dict[str, str]
                        ) -> List[PartitionInfo]:
        if 'timePartitioning' not in table:
            return []

        query = textwrap.dedent("""
        SELECT partition_id,
               TIMESTAMP(creation_time/1000) AS creation_time
        FROM   [{project}:{dataset}.{table}$__PARTITIONS_SUMMARY__]
        WHERE  partition_id <> '__UNPARTITIONED__'
               AND partition_id <> '__NULL__'
        """)
        body = {
            'query': query.format(
                project=tableRef['projectId'],
                dataset=tableRef['datasetId'],
                table=tableRef['tableId']),
            'useLegacySql': True
        }
        result = self.bigquery_service.jobs().query(projectId=self.project_id, body=body).execute()

        if 'rows' not in result:
            return []

        return [PartitionInfo(row['f'][0]['v'], row['f'][1]['v']) for row in result['rows']]

    def _get_partition_watermarks(self,
                                  table: Dict[str, Any],
                                  tableRef: Dict[str, str],
                                  partitions: List[PartitionInfo]
                                  ) -> Tuple[Watermark, Watermark]:
        if 'field' in table['timePartitioning']:
            field = table['timePartitioning']['field']
        else:
            field = '_PARTITIONTIME'

        low = min(partitions, key=lambda t: t.partition_id)
        low_wm = Watermark(
            datetime.datetime.fromtimestamp(float(low.epoch_created)).strftime('%Y-%m-%d %H:%M:%S'),
            'bigquery',
            tableRef['datasetId'],
            tableRef['tableId'],
            f'{field}={low.partition_id}',
            part_type="low_watermark",
            cluster=tableRef['projectId']
        )

        high = max(partitions, key=lambda t: t.partition_id)
        high_wm = Watermark(
            datetime.datetime.fromtimestamp(float(high.epoch_created)).strftime('%Y-%m-%d %H:%M:%S'),
            'bigquery',
            tableRef['datasetId'],
            tableRef['tableId'],
            f'{field}={high.partition_id}',
            part_type="high_watermark",
            cluster=tableRef['projectId']
        )

        return low_wm, high_wm



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

"""
The Generic Usage Extractor allows you to populate the "Frequent Users" and "Popular Tables" features within
Amundsen with the help of the TableColumnUsage class. Because this is a generic usage extractor, you need to create
a custom log parser to calculate poplarity. There is an example of how to calculate table popularity for Snowflake
in databuilder/example/scripts/sample_snowflake_table_usage.scala.
"""

import logging
from typing import (
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder import Scoped
from databuilder.extractor import sql_alchemy_extractor
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_column_usage import ColumnReader, TableColumnUsage

LOGGER = logging.getLogger(__name__)


class GenericUsageExtractor(Extractor):
    # SELECT statement from table that contains usage data
    SQL_STATEMENT = """
        SELECT
            database,
            schema,
            name,
            user_email,
            read_count
        FROM
            {database}.{schema}.{table}
        {where_clause_suffix};
    """
    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    POPULARITY_TABLE_DATABASE = 'popularity_table_database'
    POPULARTIY_TABLE_SCHEMA = 'popularity_table_schema'
    POPULARITY_TABLE_NAME = 'popularity_table_name'
    DATABASE_KEY = 'database_key'

    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {WHERE_CLAUSE_SUFFIX_KEY: ' ',
         POPULARITY_TABLE_DATABASE: 'PROD',
         POPULARTIY_TABLE_SCHEMA: 'SCHEMA',
         POPULARITY_TABLE_NAME: 'TABLE',
         DATABASE_KEY: 'snowflake'}
    )

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(GenericUsageExtractor.DEFAULT_CONFIG)
        self._where_clause_suffix = conf.get_string(GenericUsageExtractor.WHERE_CLAUSE_SUFFIX_KEY)
        self._popularity_table_database = conf.get_string(GenericUsageExtractor.POPULARITY_TABLE_DATABASE)
        self._popularity_table_schema = conf.get_string(GenericUsageExtractor.POPULARTIY_TABLE_SCHEMA)
        self._popularity_table_name = conf.get_string(GenericUsageExtractor.POPULARITY_TABLE_NAME)
        self._database_key = conf.get_string(GenericUsageExtractor.DATABASE_KEY)

        self.sql_stmt = self.SQL_STATEMENT.format(
            where_clause_suffix=self._where_clause_suffix,
            database=self._popularity_table_database,
            schema=self._popularity_table_schema,
            table=self._popularity_table_name
        )

        LOGGER.info("SQL for popularity: {}".format(self.sql_stmt))

        self._alchemy_extractor = sql_alchemy_extractor.from_surrounding_config(conf, self.sql_stmt)
        sql_alch_conf = Scoped.get_scoped_conf(conf, self._alchemy_extractor.get_scope()) \
            .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: self.sql_stmt}))
        self._alchemy_extractor.init(sql_alch_conf)
        self._extract_iter: Union[None, Iterator] = None

    def extract(self) -> Union[TableColumnUsage, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return "extractor.generic_usage"

    def _get_extract_iter(self) -> Iterator[TableColumnUsage]:
        """
        Using raw level iterator, it groups to table and yields TableColumnUsage
        :return:
        """
        for row in self._get_raw_extract_iter():
            col_readers = []
            col_readers.append(ColumnReader(database=self._database_key,
                                            cluster=row["database"],
                                            schema=row["schema"],
                                            table=row["name"],
                                            column="*",
                                            user_email=row["user_email"],
                                            read_count=row["read_count"]))
            yield TableColumnUsage(col_readers=col_readers)

    def _get_raw_extract_iter(self) -> Iterator[Dict[str, Any]]:
        """
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        """
        row = self._alchemy_extractor.extract()
        while row:
            yield row
            row = self._alchemy_extractor.extract()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


from typing import Iterator, Optional
from xml.etree import ElementTree

import requests
from pyhocon import ConfigTree
from requests.auth import HTTPBasicAuth

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.user import User


class BamboohrUserExtractor(Extractor):
    API_KEY = 'api_key'
    SUBDOMAIN = 'subdomain'

    def init(self, conf: ConfigTree) -> None:
        self._extract_iter: Optional[Iterator] = None
        self._extract_iter = None

        self._api_key = conf.get_string(BamboohrUserExtractor.API_KEY)
        self._subdomain = conf.get_string(BamboohrUserExtractor.SUBDOMAIN)

    def extract(self) -> Optional[User]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _employee_directory_uri(self) -> str:
        return f'https://api.bamboohr.com/api/gateway.php/{self._subdomain}/v1/employees/directory'

    def _get_extract_iter(self) -> Iterator[User]:
        response = requests.get(
            self._employee_directory_uri(), auth=HTTPBasicAuth(self._api_key, 'x')
        )

        root = ElementTree.fromstring(response.content)

        for user in root.findall('./employees/employee'):

            def get_field(name: str) -> str:
                field = user.find(f"./field[@id='{name}']")
                if field is not None and field.text is not None:
                    return field.text
                else:
                    return ''

            yield User(
                email=get_field('workEmail'),
                first_name=get_field('firstName'),
                last_name=get_field('lastName'),
                name=get_field('displayName'),
                team_name=get_field('department'),
                role_name=get_field('jobTitle'),
            )

    def get_scope(self) -> str:
        return 'extractor.bamboohr_user'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (  # noqa: F401
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401

from databuilder.extractor.base_teradata_metadata_extractor import BaseTeradataMetadataExtractor


class TeradataMetadataExtractor(BaseTeradataMetadataExtractor):
    """
    Extracts Teradata table and column metadata from underlying meta store database using SQLAlchemyExtractor
    """

    def get_sql_statement(
        self, use_catalog_as_cluster_name: bool, where_clause_suffix: str
    ) -> str:
        if use_catalog_as_cluster_name:
            cluster_source = "current_database()"
        else:
            cluster_source = f"'{self._cluster}'"

        return """
            SELECT
                {cluster_source} as td_cluster,
                c.DatabaseName as schema,
                c.TableName as name,
                c.CommentString as description,
                d.ColumnName as col_name,
                d.ColumnType as col_type,
                d.CommentString as col_description,
                d.ColumnId as col_sort_order
            FROM dbc.Tables c, dbc.Columns d
            WHERE c.DatabaseName = d.DatabaseName AND c.TableName = d.TableName
            AND {where_clause_suffix}
            ORDER by cluster_a, schema, name, col_sort_order;
        """.format(
            cluster_source=cluster_source,
            where_clause_suffix=where_clause_suffix,
        )

    def get_scope(self) -> str:
        return "extractor.teradata_metadata"



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Dict, Iterator, Optional, Union,
)

from databuilder.models.feature.feature_metadata import FeatureMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable


# FeatureGenerationCode allows ingesting as text the generation code - whether sql or not -
# which was used to create a feature. Unlike the Query node for Dashboards, it has no inherent
# concept of name, url, id, or hierarchical structure. This allows for maximum flexibility to
# ingest generation code regardless of source.
class FeatureGenerationCode(GraphSerializable):
    NODE_LABEL = 'Feature_Generation_Code'

    TEXT_ATTR = 'text'
    LAST_EXECUTED_TIMESTAMP_ATTR = 'last_executed_timestamp'
    SOURCE_ATTR = 'source'

    FEATURE_GENCODE_RELATION_TYPE = 'GENERATION_CODE'
    GENCODE_FEATURE_RELATION_TYPE = 'GENERATION_CODE_OF'

    def __init__(self,
                 feature_group: str,
                 feature_name: str,
                 feature_version: str,
                 text: str,
                 source: Optional[str] = None,
                 last_executed_timestamp: Optional[int] = None,
                 **kwargs: Any
                 ) -> None:

        self.feature_group = feature_group
        self.feature_name = feature_name
        self.feature_version = feature_version
        self.text = text
        self.source = source
        self.last_executed_timestamp = last_executed_timestamp

        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()

    def __repr__(self) -> str:
        return f'Feature_Generation_Code({self.feature_group!r}, {self.feature_name!r}, {self.feature_version!r}, ' \
               f'{self.text!r}, {self.source!r}, {self.last_executed_timestamp!r})'

    def _get_feature_key(self) -> str:
        return FeatureMetadata.KEY_FORMAT.format(feature_group=self.feature_group,
                                                 name=self.feature_name,
                                                 version=self.feature_version)

    def _get_generation_code_key(self) -> str:
        return f'{self._get_feature_key()}/_generation_code'

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        attrs: Dict[str, Any] = {
            FeatureGenerationCode.TEXT_ATTR: self.text,
        }
        if self.last_executed_timestamp:
            attrs[FeatureGenerationCode.LAST_EXECUTED_TIMESTAMP_ATTR] = self.last_executed_timestamp

        if self.source:
            attrs[FeatureGenerationCode.SOURCE_ATTR] = self.source

        yield GraphNode(
            key=self._get_generation_code_key(),
            label=FeatureGenerationCode.NODE_LABEL,
            attributes=attrs,
        )

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=FeatureMetadata.NODE_LABEL,
            end_label=FeatureGenerationCode.NODE_LABEL,
            start_key=self._get_feature_key(),
            end_key=self._get_generation_code_key(),
            type=FeatureGenerationCode.FEATURE_GENCODE_RELATION_TYPE,
            reverse_type=FeatureGenerationCode.GENCODE_FEATURE_RELATION_TYPE,
            attributes={},
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import csv
import ctypes
import logging
import time
from io import open
from typing import (
    Dict, List, Set,
)

import neo4j
import pandas
from jinja2 import Template
from neo4j import GraphDatabase, Neo4jDriver
from neo4j.api import (
    SECURITY_TYPE_SECURE, SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, parse_neo4j_uri,
)
from pyhocon import ConfigFactory, ConfigTree

from databuilder.models.graph_serializable import (
    NODE_KEY, NODE_LABEL, RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY,
    RELATION_START_LABEL, RELATION_TYPE,
)
from databuilder.publisher.base_publisher import Publisher
from databuilder.publisher.publisher_config_constants import (
    Neo4jCsvPublisherConfigs, PublishBehaviorConfigs, PublisherConfigs,
)
from databuilder.utils.publisher_utils import (
    chunkify_list, create_neo4j_node_key_constraint, create_props_param, execute_neo4j_statement, get_props_body_keys,
    list_files,
)

# Setting field_size_limit to solve the error below
# _csv.Error: field larger than field limit (131072)
# https://stackoverflow.com/a/54517228/5972935
csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))

# Required columns for Node
NODE_REQUIRED_KEYS = {NODE_LABEL, NODE_KEY}
# Required columns for Relationship
RELATION_REQUIRED_KEYS = {RELATION_START_LABEL, RELATION_START_KEY,
                          RELATION_END_LABEL, RELATION_END_KEY,
                          RELATION_TYPE, RELATION_REVERSE_TYPE}

DEFAULT_CONFIG = ConfigFactory.from_dict({Neo4jCsvPublisherConfigs.NEO4J_TRANSACTION_SIZE: 1000,
                                          Neo4jCsvPublisherConfigs.NEO4J_MAX_CONN_LIFE_TIME_SEC: 50,
                                          Neo4jCsvPublisherConfigs.NEO4J_DATABASE_NAME: neo4j.DEFAULT_DATABASE,
                                          PublishBehaviorConfigs.ADD_PUBLISHER_METADATA: True,
                                          PublishBehaviorConfigs.PUBLISH_REVERSE_RELATIONSHIPS: True,
                                          PublishBehaviorConfigs.PRESERVE_ADHOC_UI_DATA: True,
                                          PublishBehaviorConfigs.PRESERVE_EMPTY_PROPS: True})

LOGGER = logging.getLogger(__name__)


class Neo4jCsvUnwindPublisher(Publisher):
    """
    This publisher takes two folders for input and publishes to Neo4j.
    One folder will contain CSV file(s) for Nodes where the other folder will contain CSV
    file(s) for Relationships.

    The merge statements make use of the UNWIND clause to allow for batched params to be applied to each
    statement. This improves performance by reducing the amount of individual transactions to the database,
    and by allowing Neo4j to compile and cache the statement.
    """

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(DEFAULT_CONFIG)

        self._count: int = 0
        self._node_files = list_files(conf, PublisherConfigs.NODE_FILES_DIR)
        self._node_files_iter = iter(self._node_files)

        self._relation_files = list_files(conf, PublisherConfigs.RELATION_FILES_DIR)
        self._relation_files_iter = iter(self._relation_files)

        self._driver = self._driver_init(conf)
        self._db_name = conf.get_string(Neo4jCsvPublisherConfigs.NEO4J_DATABASE_NAME)
        self._transaction_size = conf.get_int(Neo4jCsvPublisherConfigs.NEO4J_TRANSACTION_SIZE)

        # config is list of node label.
        # When set, this list specifies a list of nodes that shouldn't be updated, if exists
        self._create_only_nodes = set(conf.get_list(Neo4jCsvPublisherConfigs.NEO4J_CREATE_ONLY_NODES, default=[]))
        self._labels: Set[str] = set()
        self._publish_tag: str = conf.get_string(PublisherConfigs.JOB_PUBLISH_TAG)
        self._additional_publisher_metadata_fields: Dict =\
            dict(conf.get(PublisherConfigs.ADDITIONAL_PUBLISHER_METADATA_FIELDS, default={}))
        self._add_publisher_metadata: bool = conf.get_bool(PublishBehaviorConfigs.ADD_PUBLISHER_METADATA)
        self._publish_reverse_relationships: bool = conf.get_bool(PublishBehaviorConfigs.PUBLISH_REVERSE_RELATIONSHIPS)
        self._preserve_adhoc_ui_data = conf.get_bool(PublishBehaviorConfigs.PRESERVE_ADHOC_UI_DATA)
        self._preserve_empty_props: bool = conf.get_bool(PublishBehaviorConfigs.PRESERVE_EMPTY_PROPS)
        if self._add_publisher_metadata and not self._publish_tag:
            raise Exception(f'{PublisherConfigs.JOB_PUBLISH_TAG} should not be empty')

        LOGGER.info('Publishing Node csv files %s, and Relation CSV files %s',
                    self._node_files,
                    self._relation_files)

    def _driver_init(self, conf: ConfigTree) -> Neo4jDriver:
        uri = conf.get_string(Neo4jCsvPublisherConfigs.NEO4J_END_POINT_KEY)
        driver_args = {
            'uri': uri,
            'max_connection_lifetime': conf.get_int(Neo4jCsvPublisherConfigs.NEO4J_MAX_CONN_LIFE_TIME_SEC),
            'auth': (conf.get_string(Neo4jCsvPublisherConfigs.NEO4J_USER),
                     conf.get_string(Neo4jCsvPublisherConfigs.NEO4J_PASSWORD)),
        }

        # if URI scheme not secure set `trust`` and `encrypted` to default values
        # https://neo4j.com/docs/api/python-driver/current/api.html#uri
        _, security_type, _ = parse_neo4j_uri(uri=uri)
        if security_type not in [SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, SECURITY_TYPE_SECURE]:
            default_security_conf = {'trust': neo4j.TRUST_ALL_CERTIFICATES, 'encrypted': True}
            driver_args.update(default_security_conf)

        # if NEO4J_VALIDATE_SSL or NEO4J_ENCRYPTED are set in config pass them to the driver
        validate_ssl_conf = conf.get(Neo4jCsvPublisherConfigs.NEO4J_VALIDATE_SSL, None)
        encrypted_conf = conf.get(Neo4jCsvPublisherConfigs.NEO4J_ENCRYPTED, None)
        if validate_ssl_conf is not None:
            driver_args['trust'] = neo4j.TRUST_SYSTEM_CA_SIGNED_CERTIFICATES if validate_ssl_conf \
                else neo4j.TRUST_ALL_CERTIFICATES
        if encrypted_conf is not None:
            driver_args['encrypted'] = encrypted_conf

        driver = GraphDatabase.driver(**driver_args)

        try:
            driver.verify_connectivity()
        except Exception as e:
            driver.close()
            raise e

        return driver

    def publish_impl(self) -> None:  # noqa: C901
        """
        Publishes Nodes first and then Relations
        """
        start = time.time()

        for node_file in self._node_files:
            self.pre_publish_node_file(node_file)

        LOGGER.info('Publishing Node files: %s', self._node_files)
        while True:
            try:
                node_file = next(self._node_files_iter)
                self._publish_node_file(node_file)
            except StopIteration:
                break

        for rel_file in self._relation_files:
            self.pre_publish_rel_file(rel_file)

        LOGGER.info('Publishing Relationship files: %s', self._relation_files)
        while True:
            try:
                relation_file = next(self._relation_files_iter)
                self._publish_relation_file(relation_file)
            except StopIteration:
                break

        LOGGER.info('Committed total %i statements', self._count)

        # TODO: Add statsd support
        LOGGER.info('Successfully published. Elapsed: %i seconds', time.time() - start)

    def get_scope(self) -> str:
        return 'publisher.neo4j'

    # Can be overridden with custom action(s)
    def pre_publish_node_file(self, node_file: str) -> None:
        created_constraint_labels = create_neo4j_node_key_constraint(node_file, self._labels,
                                                                     self._driver, self._db_name)
        self._labels.union(created_constraint_labels)

    # Can be overridden with custom action(s)
    def pre_publish_rel_file(self, rel_file: str) -> None:
        pass

    def _publish_node_file(self, node_file: str) -> None:
        with open(node_file, 'r', encoding='utf8') as node_csv:
            csv_dataframe = pandas.read_csv(node_csv, na_filter=False)
            all_node_records = csv_dataframe.to_dict(orient="records")

            # Get the first node label since they will be the same for all records in the file
            merge_stmt = self._create_node_merge_statement(node_keys=csv_dataframe.columns.tolist(),
                                                           node_label=all_node_records[0][NODE_LABEL])

            self._write_transactions(merge_stmt, all_node_records)

    def _create_node_merge_statement(self, node_keys: list, node_label: str) -> str:
        template = Template("""
            UNWIND $batch AS row
            MERGE (node:{{ LABEL }} {key: row.KEY})
            ON CREATE SET {{ PROPS_BODY_CREATE }}
            {% if update %} ON MATCH SET {{ PROPS_BODY_UPDATE }} {% endif %}
        """)

        props_body_create = self._create_props_body(get_props_body_keys(node_keys,
                                                    NODE_REQUIRED_KEYS,
                                                    self._additional_publisher_metadata_fields), 'node')

        props_body_update = props_body_create
        if self._preserve_adhoc_ui_data:
            props_body_update = self._create_props_body(get_props_body_keys(node_keys,
                                                                            NODE_REQUIRED_KEYS,
                                                                            self._additional_publisher_metadata_fields),
                                                        'node', True)

        return template.render(LABEL=node_label,
                               PROPS_BODY_CREATE=props_body_create,
                               PROPS_BODY_UPDATE=props_body_update,
                               update=(node_label not in self._create_only_nodes))

    def _publish_relation_file(self, relation_file: str) -> None:
        with open(relation_file, 'r', encoding='utf8') as relation_csv:
            csv_dataframe = pandas.read_csv(relation_csv, na_filter=False)
            all_rel_records = csv_dataframe.to_dict(orient="records")

            # Get the first relation labels since they will be the same for all records in the file
            merge_stmt = self._create_relationship_merge_statement(
                rel_keys=csv_dataframe.columns.tolist(),
                start_label=all_rel_records[0][RELATION_START_LABEL],
                end_label=all_rel_records[0][RELATION_END_LABEL],
                relation_type=all_rel_records[0][RELATION_TYPE],
                relation_reverse_type=all_rel_records[0][RELATION_REVERSE_TYPE]
            )

            self._write_transactions(merge_stmt, all_rel_records)

    def _create_relationship_merge_statement(self,
                                             rel_keys: list,
                                             start_label: str,
                                             end_label: str,
                                             relation_type: str,
                                             relation_reverse_type: str) -> str:
        template = Template("""
            UNWIND $batch as row
            MATCH (n1:{{ START_LABEL }} {key: row.START_KEY}), (n2:{{ END_LABEL }} {key: row.END_KEY})
            {% if publish_reverse_relationships %}
            MERGE (n1)-[r1:{{ TYPE }}]->(n2)-[r2:{{ REVERSE_TYPE }}]->(n1)
            {% elif not publish_reverse_relationships and has_key %}
            MERGE (n1)-[r1:{{ TYPE }} {key: row.key}]->(n2)
            {% else %}
            MERGE (n1)-[r1:{{ TYPE }}]->(n2)
            {% endif %}
            {% if update_props_body %}
            ON CREATE SET {{ props_body_create }}
            ON MATCH SET {{ props_body_update }}
            {% endif %}
            RETURN n1.key, n2.key
        """)

        props_body_template = Template("""{{ props_body_r1 }} , {{ props_body_r2 }}""")

        props_body_r1 = self._create_props_body(get_props_body_keys(rel_keys,
                                                                    RELATION_REQUIRED_KEYS,
                                                                    self._additional_publisher_metadata_fields), 'r1')
        props_body_r2 = self._create_props_body(get_props_body_keys(rel_keys,
                                                                    RELATION_REQUIRED_KEYS,
                                                                    self._additional_publisher_metadata_fields), 'r2')
        if self._publish_reverse_relationships:
            props_body_create = props_body_template.render(props_body_r1=props_body_r1, props_body_r2=props_body_r2)
        else:
            props_body_create = props_body_r1

        props_body_update = props_body_create
        if self._preserve_adhoc_ui_data:
            props_body_r1 = self._create_props_body(get_props_body_keys(rel_keys,
                                                                        RELATION_REQUIRED_KEYS,
                                                                        self._additional_publisher_metadata_fields),
                                                    'r1', True)
            props_body_r2 = self._create_props_body(get_props_body_keys(rel_keys,
                                                                        RELATION_REQUIRED_KEYS,
                                                                        self._additional_publisher_metadata_fields),
                                                    'r2', True)
            if self._publish_reverse_relationships:
                props_body_update = props_body_template.render(props_body_r1=props_body_r1, props_body_r2=props_body_r2)
            else:
                props_body_update = props_body_r1

        return template.render(START_LABEL=start_label,
                               END_LABEL=end_label,
                               publish_reverse_relationships=self._publish_reverse_relationships,
                               has_key='key' in rel_keys,
                               TYPE=relation_type,
                               REVERSE_TYPE=relation_reverse_type,
                               update_props_body=props_body_r1,
                               props_body_create=props_body_create,
                               props_body_update=props_body_update)

    def _create_props_body(self,
                           record_keys: Set,
                           identifier: str,
                           rename_id_to_preserve_ui_data: bool = False) -> str:
        """
        Creates properties body with params required for resolving template.

        e.g: Note that node.key3 is not quoted if header has UNQUOTED_SUFFIX.
        identifier.key1 = 'val1' , identifier.key2 = 'val2', identifier.key3 = val3

        :param record_keys: a list of keys for a CSV row
        :param identifier: identifier that will be used in CYPHER query as shown on above example
        :param rename_id_to_preserve_ui_data: specifies whether to null out the identifier to prevent it from updating
        :return: Properties body for Cypher statement
        """
        # For SET, if the evaluated expression is null, no action is performed. I.e. `SET (null).foo = 5` is a noop.
        # See https://neo4j.com/docs/cypher-manual/current/clauses/set/
        if rename_id_to_preserve_ui_data:
            identifier = f"""
                (CASE WHEN {identifier}.{PublisherConfigs.PUBLISHED_TAG_PROPERTY_NAME} IS NOT NULL
                THEN {identifier} ELSE null END)
            """

        template = Template("""
            {% for k in record_keys %}
                {% if preserve_empty_props %}
                    {{ identifier }}.{{ k }} = row.{{ k }}
                {% else %}
                    {{ identifier }}.{{ k }} = (CASE row.{{ k }} WHEN '' THEN NULL ELSE row.{{ k }} END)
                {% endif %}
                {{ ", " if not loop.last else "" }}
            {% endfor %}
            {% if record_keys and add_publisher_metadata %}
                ,
            {% endif %}
            {% if add_publisher_metadata %}
                {{ identifier }}.{{ published_tag_prop }} = '{{ publish_tag }}',
                {{ identifier }}.{{ last_updated_prop }} = timestamp()
            {% endif %}
        """)

        props_body = template.render(record_keys=record_keys,
                                     preserve_empty_props=self._preserve_empty_props,
                                     identifier=identifier,
                                     add_publisher_metadata=self._add_publisher_metadata,
                                     published_tag_prop=PublisherConfigs.PUBLISHED_TAG_PROPERTY_NAME,
                                     publish_tag=self._publish_tag,
                                     last_updated_prop=PublisherConfigs.LAST_UPDATED_EPOCH_MS)
        return props_body.strip()

    def _write_transactions(self,
                            stmt: str,
                            records: List[dict]) -> None:
        for chunk in chunkify_list(records, self._transaction_size):
            params_list = []
            for record in chunk:
                params_list.append(create_props_param(record, self._additional_publisher_metadata_fields))

            with self._driver.session(database=self._db_name) as session:
                session.write_transaction(execute_neo4j_statement, stmt, {'batch': params_list})

                self._count += len(params_list)
                LOGGER.info(f'Committed {self._count} rows so far')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import textwrap
import time
from typing import (
    Any, Dict, Iterable, Union,
)

import neo4j
from neo4j import GraphDatabase
from neo4j.api import (
    SECURITY_TYPE_SECURE, SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, parse_neo4j_uri,
)
from pyhocon import ConfigFactory, ConfigTree

from databuilder import Scoped
from databuilder.publisher.neo4j_csv_publisher import JOB_PUBLISH_TAG
from databuilder.task.base_task import Task

# A end point for Neo4j e.g: bolt://localhost:9999
NEO4J_END_POINT_KEY = 'neo4j_endpoint'
NEO4J_MAX_CONN_LIFE_TIME_SEC = 'neo4j_max_conn_life_time_sec'
NEO4J_USER = 'neo4j_user'
NEO4J_PASSWORD = 'neo4j_password'
# in Neo4j (v4.0+), we can create and use more than one active database at the same time
NEO4J_DATABASE_NAME = 'neo4j_database'
NEO4J_ENCRYPTED = 'neo4j_encrypted'
"""NEO4J_ENCRYPTED is a boolean indicating whether to use SSL/TLS when connecting."""
NEO4J_VALIDATE_SSL = 'neo4j_validate_ssl'
"""NEO4J_VALIDATE_SSL is a boolean indicating whether to validate the server's SSL/TLS cert against system CAs."""
TARGET_NODES = "target_nodes"
TARGET_RELATIONS = "target_relations"
BATCH_SIZE = "batch_size"
DRY_RUN = "dry_run"
# Staleness max percentage. Safety net to prevent majority of data being deleted.
STALENESS_MAX_PCT = "staleness_max_pct"
# Staleness max percentage per LABEL/TYPE. Safety net to prevent majority of data being deleted.
STALENESS_PCT_MAX_DICT = "staleness_max_pct_dict"
# Using this milliseconds and published timestamp to determine staleness
MS_TO_EXPIRE = "milliseconds_to_expire"
MIN_MS_TO_EXPIRE = "minimum_milliseconds_to_expire"
RETAIN_DATA_WITH_NO_PUBLISHER_METADATA = "retain_data_with_no_publisher_metadata"

DEFAULT_CONFIG = ConfigFactory.from_dict({BATCH_SIZE: 100,
                                          NEO4J_MAX_CONN_LIFE_TIME_SEC: 50,
                                          NEO4J_DATABASE_NAME: neo4j.DEFAULT_DATABASE,
                                          STALENESS_MAX_PCT: 5,
                                          TARGET_NODES: [],
                                          TARGET_RELATIONS: [],
                                          STALENESS_PCT_MAX_DICT: {},
                                          MIN_MS_TO_EXPIRE: 86400000,
                                          RETAIN_DATA_WITH_NO_PUBLISHER_METADATA: False,
                                          DRY_RUN: False})

LOGGER = logging.getLogger(__name__)

MARKER_VAR_NAME = 'marker'


class TargetWithCondition:
    def __init__(self, target_type: str, condition: str) -> None:
        self.target_type = target_type
        self.condition = condition


class Neo4jStalenessRemovalTask(Task):
    """
    A Specific task that is to remove stale nodes and relations in Neo4j.
    It will use "published_tag" attribute assigned from Neo4jCsvPublisher and if "published_tag" is different from
    the one it is getting it from the config, it will regard the node/relation as stale.
    Not all resource is being published by Neo4jCsvPublisher and you can only set specific LABEL of the node or TYPE
    of relation to perform this deletion.

    """

    delete_stale_nodes_statement = textwrap.dedent("""
        MATCH (target:{{type}})
        WHERE {staleness_condition}{{extra_condition}}
        WITH target LIMIT $batch_size
        DETACH DELETE (target)
        RETURN count(*) as count
        """)
    delete_stale_relations_statement = textwrap.dedent("""
        MATCH (start_node)-[target:{{type}}]-(end_node)
        WHERE {staleness_condition}{{extra_condition}}
        WITH target LIMIT $batch_size
        DELETE target
        RETURN count(*) as count
        """)
    validate_node_staleness_statement = textwrap.dedent("""
        MATCH (target:{{type}})
        WHERE {staleness_condition}{{extra_condition}}
        RETURN count(*) as count
        """)
    validate_relation_staleness_statement = textwrap.dedent("""
        MATCH (start_node)-[target:{{type}}]-(end_node)
        WHERE {staleness_condition}{{extra_condition}}
        RETURN count(*) as count
        """)

    def __init__(self) -> None:
        pass

    def get_scope(self) -> str:
        return 'task.remove_stale_data'

    def init(self, conf: ConfigTree) -> None:
        conf = Scoped.get_scoped_conf(conf, self.get_scope()) \
            .with_fallback(conf) \
            .with_fallback(DEFAULT_CONFIG)
        self.target_nodes = set(conf.get_list(TARGET_NODES))
        self.target_relations = set(conf.get_list(TARGET_RELATIONS))
        self.batch_size = conf.get_int(BATCH_SIZE)
        self.dry_run = conf.get_bool(DRY_RUN)
        self.staleness_pct = conf.get_int(STALENESS_MAX_PCT)
        self.staleness_pct_dict = conf.get(STALENESS_PCT_MAX_DICT)
        self.retain_data_with_no_publisher_metadata = conf.get_bool(RETAIN_DATA_WITH_NO_PUBLISHER_METADATA)

        if JOB_PUBLISH_TAG in conf and MS_TO_EXPIRE in conf:
            raise Exception(f'Cannot have both {JOB_PUBLISH_TAG} and {MS_TO_EXPIRE} in job config')

        self.ms_to_expire = None
        if MS_TO_EXPIRE in conf:
            self.ms_to_expire = conf.get_int(MS_TO_EXPIRE)
            if self.ms_to_expire < conf.get_int(MIN_MS_TO_EXPIRE):
                raise Exception(f'{MS_TO_EXPIRE} is too small')
            self.marker = self.ms_to_expire
        else:
            self.marker = conf.get_string(JOB_PUBLISH_TAG)

        uri = conf.get_string(NEO4J_END_POINT_KEY)
        driver_args = {
            'uri': uri,
            'max_connection_lifetime': conf.get_int(NEO4J_MAX_CONN_LIFE_TIME_SEC),
            'auth': (conf.get_string(NEO4J_USER), conf.get_string(NEO4J_PASSWORD)),
        }

        # if URI scheme not secure set `trust`` and `encrypted` to default values
        # https://neo4j.com/docs/api/python-driver/current/api.html#uri
        _, security_type, _ = parse_neo4j_uri(uri=uri)
        if security_type not in [SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, SECURITY_TYPE_SECURE]:
            default_security_conf = {'trust': neo4j.TRUST_ALL_CERTIFICATES, 'encrypted': True}
            driver_args.update(default_security_conf)

        # if NEO4J_VALIDATE_SSL or NEO4J_ENCRYPTED are set in config pass them to the driver
        validate_ssl_conf = conf.get(NEO4J_VALIDATE_SSL, None)
        encrypted_conf = conf.get(NEO4J_ENCRYPTED, None)
        if validate_ssl_conf is not None:
            driver_args['trust'] = neo4j.TRUST_SYSTEM_CA_SIGNED_CERTIFICATES if validate_ssl_conf \
                else neo4j.TRUST_ALL_CERTIFICATES
        if encrypted_conf is not None:
            driver_args['encrypted'] = encrypted_conf

        self._driver = GraphDatabase.driver(**driver_args)

        self.db_name = conf.get(NEO4J_DATABASE_NAME)

    def run(self) -> None:
        """
        First, performs a safety check to make sure this operation would not delete more than a threshold where
        default threshold is 5%. Once it passes a safety check, it will first delete stale nodes, and then stale
        relations.
        :return:
        """
        self.validate()
        self._delete_stale_nodes()
        self._delete_stale_relations()

    def validate(self) -> None:
        """
        Validation method. Focused on limit the risk on deleting nodes and relations.
         - Check if deleted nodes will be within 10% of total nodes.
        :return:
        """
        self._validate_node_staleness_pct()
        self._validate_relation_staleness_pct()

    def _delete_stale_nodes(self) -> None:
        self._batch_delete(statement=self._decorate_staleness(self.delete_stale_nodes_statement),
                           targets=self.target_nodes)

    def _decorate_staleness(self,
                            statement: str
                            ) -> str:
        """
        Append where clause to the Cypher statement depends on which field to be used to expire stale data.
        :param statement:
        :return:
        """
        if self.ms_to_expire:
            condition = f"""target.publisher_last_updated_epoch_ms < (timestamp() - ${MARKER_VAR_NAME})"""
            if not self.retain_data_with_no_publisher_metadata:
                null_check = "EXISTS(target.publisher_last_updated_epoch_ms)"
                condition = f"""{condition}\nOR NOT {null_check}"""
            condition = f"""({condition})"""
            return statement.format(staleness_condition=condition)

        condition = f"""target.published_tag < ${MARKER_VAR_NAME}"""
        if not self.retain_data_with_no_publisher_metadata:
            null_check = "EXISTS(target.published_tag)"
            condition = f"""{condition}\nOR NOT {null_check}"""
        condition = f"""({condition})"""
        return statement.format(staleness_condition=condition)

    def _delete_stale_relations(self) -> None:
        self._batch_delete(statement=self._decorate_staleness(self.delete_stale_relations_statement),
                           targets=self.target_relations)

    def _batch_delete(self,
                      statement: str,
                      targets: Union[Iterable[str], Iterable[TargetWithCondition]]
                      ) -> None:
        """
        Performing huge amount of deletion could degrade Neo4j performance. Therefore, it's taking batch deletion here.
        :param statement:
        :param targets:
        :return:
        """
        for t in targets:
            if isinstance(t, TargetWithCondition):
                target_type = t.target_type
                extra_condition = ' AND ' + t.condition
            else:
                target_type = t
                extra_condition = ''

            LOGGER.info('Deleting stale data of %s with batch size %i', target_type, self.batch_size)
            total_count = 0
            while True:
                results = self._execute_cypher_query(statement=statement.format(type=target_type,
                                                                                extra_condition=extra_condition),
                                                     param_dict={'batch_size': self.batch_size,
                                                                 MARKER_VAR_NAME: self.marker},
                                                     dry_run=self.dry_run)
                record = next(iter(results), None)
                count = record['count'] if record else 0
                total_count = total_count + count
                if count == 0:
                    break
            LOGGER.info('Deleted %i stale data of %s', total_count, target_type)

    def _validate_staleness_pct(self,
                                total_record_count: int,
                                stale_record_count: int,
                                target_type: str
                                ) -> None:
        if total_record_count == 0 or stale_record_count == 0:
            return

        stale_pct = stale_record_count * 100 / total_record_count

        threshold = self.staleness_pct_dict.get(target_type, self.staleness_pct)
        if stale_pct >= threshold:
            raise Exception(f'Staleness percentage of {target_type} is {stale_pct} %. '
                            f'Stopping due to over threshold {threshold} %')

    def _validate_node_staleness_pct(self) -> None:
        total_nodes_statement = textwrap.dedent(
            self.validate_node_staleness_statement.format(staleness_condition='true'))
        stale_nodes_statement = textwrap.dedent(
            self._decorate_staleness(self.validate_node_staleness_statement))

        for t in self.target_nodes:
            if isinstance(t, TargetWithCondition):
                target_type = t.target_type
                extra_condition = ' AND ' + t.condition
            else:
                target_type = t
                extra_condition = ''

            total_records = self._execute_cypher_query(
                statement=total_nodes_statement.format(type=target_type,
                                                       extra_condition=extra_condition))
            stale_records = self._execute_cypher_query(
                statement=stale_nodes_statement.format(type=target_type,
                                                       extra_condition=extra_condition),
                param_dict={MARKER_VAR_NAME: self.marker})

            total_record_value = next(iter(total_records), None)
            stale_record_value = next(iter(stale_records), None)
            self._validate_staleness_pct(total_record_count=total_record_value['count'] if total_record_value else 0,
                                         stale_record_count=stale_record_value['count'] if stale_record_value else 0,
                                         target_type=target_type)

    def _validate_relation_staleness_pct(self) -> None:
        total_relations_statement = textwrap.dedent(
            self.validate_relation_staleness_statement.format(staleness_condition='true'))
        stale_relations_statement = textwrap.dedent(
            self._decorate_staleness(self.validate_relation_staleness_statement))

        for t in self.target_relations:
            if isinstance(t, TargetWithCondition):
                target_type = t.target_type
                extra_condition = ' AND ' + t.condition
            else:
                target_type = t
                extra_condition = ''

            total_records = self._execute_cypher_query(
                statement=total_relations_statement.format(type=target_type,
                                                           extra_condition=extra_condition))
            stale_records = self._execute_cypher_query(
                statement=stale_relations_statement.format(type=target_type,
                                                           extra_condition=extra_condition),
                param_dict={MARKER_VAR_NAME: self.marker})

            total_record_value = next(iter(total_records), None)
            stale_record_value = next(iter(stale_records), None)
            self._validate_staleness_pct(total_record_count=total_record_value['count'] if total_record_value else 0,
                                         stale_record_count=stale_record_value['count'] if stale_record_value else 0,
                                         target_type=target_type)

    def _execute_cypher_query(self,
                              statement: str,
                              param_dict: Dict[str, Any] = {},
                              dry_run: bool = False
                              ) -> Iterable[Dict[str, Any]]:
        LOGGER.info('Executing Cypher query: %s with params %s: ', statement, param_dict)

        if dry_run:
            LOGGER.info('Skipping for it is a dryrun')
            return []

        start = time.time()
        try:
            with self._driver.session(database=self.db_name) as session:
                result = session.run(statement, **param_dict)
                return [record for record in result]

        finally:
            LOGGER.debug('Cypher query execution elapsed for %i seconds', time.time() - start)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from .query import QueryMetadata  # noqa
from .query_execution import QueryExecutionsMetadata  # noqa
from .query_join import QueryJoinMetadata  # noqa
from .query_where import QueryWhereMetadata  # noqa



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, List, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasCommonTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardOwner as RDSDashboardOwner
from amundsen_rds.models.table import TableOwner as RDSTableOwner
from amundsen_rds.models.user import User as RDSUser

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.owner_constants import OWNER_OF_OBJECT_RELATION_TYPE, OWNER_RELATION_TYPE
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.user import User
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


class Owner(GraphSerializable, TableSerializable, AtlasSerializable):
    LABELS_PERMITTED_TO_HAVE_OWNER = ['Table', 'Dashboard', 'Feature']

    def __init__(self,
                 start_label: str,
                 start_key: str,
                 owner_emails: Union[List, str],
                 ) -> None:
        if start_label not in Owner.LABELS_PERMITTED_TO_HAVE_OWNER:
            raise Exception(f'owners for {start_label} are not supported')
        self.start_label = start_label
        self.start_key = start_key
        if isinstance(owner_emails, str):
            owner_emails = owner_emails.split(',')
        self.owner_emails = [email.strip().lower() for email in owner_emails]

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def __repr__(self) -> str:
        return f'Owner({self.start_label!r}, {self.start_key!r}, {self.owner_emails!r})'

    def create_next_node(self) -> Optional[GraphNode]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        for email in self.owner_emails:
            if email:
                yield GraphNode(
                    key=User.get_user_model_key(email=email),
                    label=User.USER_NODE_LABEL,
                    attributes={
                        User.USER_NODE_EMAIL: email,
                    }
                )

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        for email in self.owner_emails:
            if email:
                yield GraphRelationship(
                    start_label=self.start_label,
                    start_key=self.start_key,
                    end_label=User.USER_NODE_LABEL,
                    end_key=User.get_user_model_key(email=email),
                    type=OWNER_RELATION_TYPE,
                    reverse_type=OWNER_OF_OBJECT_RELATION_TYPE,
                    attributes={}
                )

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        for email in self.owner_emails:
            if email:
                user_record = RDSUser(
                    rk=User.get_user_model_key(email=email),
                    email=email
                )
                yield user_record

                if self.start_label == TableMetadata.TABLE_NODE_LABEL:
                    yield RDSTableOwner(
                        table_rk=self.start_key,
                        user_rk=User.get_user_model_key(email=email),
                    )
                elif self.start_label == DashboardMetadata.DASHBOARD_NODE_LABEL:
                    yield RDSDashboardOwner(
                        dashboard_rk=self.start_key,
                        user_rk=User.get_user_model_key(email=email)
                    )
                else:
                    raise Exception(f'{self.start_label}<>Owner relationship is not table serializable')

    def _create_atlas_owner_entity(self, owner: str) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, owner),
            ('email', owner)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.user,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def _create_atlas_owner_relation(self, owner: str) -> AtlasRelationship:
        table_relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.resource_owner,
            entityType1=AtlasCommonTypes.data_set,
            entityQualifiedName1=self.start_key,
            entityType2=AtlasCommonTypes.user,
            entityQualifiedName2=User.get_user_model_key(email=owner),
            attributes={}
        )

        return table_relationship

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        for owner in self.owner_emails:
            if owner:
                yield self._create_atlas_owner_entity(owner)

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        for owner in self.owner_emails:
            if owner:
                yield self._create_atlas_owner_relation(owner)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import logging
from asyncio.log import logger
from typing import (
    Any, Dict, Iterator, List, Optional, Union,
)

from pyhocon import ConfigTree
from schema_registry.client import Auth, SchemaRegistryClient
from schema_registry.client.utils import SchemaVersion

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

LOGGER = logging.getLogger(__name__)


class KafkaSchemaRegistryExtractor(Extractor):
    """
    Extracts the latest version of all schemas from a given
    Kafka Schema Registry URL
    """

    REGISTRY_URL_KEY = "registry_url"
    REGISTRY_USERNAME_KEY = "registry_username"
    REGISTRY_PASSWORD_KEY = "registry_password"

    def init(self, conf: ConfigTree) -> None:
        self._registry_base_url = conf.get(
            KafkaSchemaRegistryExtractor.REGISTRY_URL_KEY
        )

        self._registry_username = conf.get(
            KafkaSchemaRegistryExtractor.REGISTRY_USERNAME_KEY, None
        )

        self._registry_password = conf.get(
            KafkaSchemaRegistryExtractor.REGISTRY_PASSWORD_KEY, None
        )

        # Add authentication if user and password are provided
        if all((self._registry_username, self._registry_password)):
            self._client = SchemaRegistryClient(
                url=self._registry_base_url,
                auth=Auth(
                    username=self._registry_username,
                    password=self._registry_password
                )
            )
        else:
            self._client = SchemaRegistryClient(
                url=self._registry_base_url,
            )

        self._extract_iter: Union[None, Iterator] = None

    def extract(self) -> Union[TableMetadata, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None
        except Exception as e:
            logger.error(f'Failed to generate next table: {e}')
            return None

    def get_scope(self) -> str:
        return 'extractor.kafka_schema_registry'

    def _get_extract_iter(self) -> Optional[Iterator[TableMetadata]]:
        """
        Return an iterator generating TableMetadata for all of the schemas.
        """
        for schema_version in self._get_raw_extract_iter():
            subject = schema_version.subject
            schema = schema_version.schema.raw_schema
            LOGGER.info((f'Subject: {subject}, '
                         f'Schema: {schema}'))

            try:
                yield KafkaSchemaRegistryExtractor._create_table(
                    schema=schema,
                    subject_name=subject,
                    cluster_name=schema.get(
                        'namespace', 'kafka-schema-registry'
                    ),
                    schema_name=schema.get('name', ''),
                    schema_description=schema.get('doc', None),
                )
            except Exception as e:
                logger.warning(f'Failed to generate table for {subject}: {e}')
                continue

    def _get_raw_extract_iter(self) -> Iterator[SchemaVersion]:
        """
        Return iterator of results row from schema registry
        """
        subjects = self._client.get_subjects()

        LOGGER.info(f'Number of extracted subjects: {len(subjects)}')
        LOGGER.info(f'Extracted subjects: {subjects}')

        for subj in subjects:
            subj_schema = self._client.get_schema(subj)
            LOGGER.info(f'Subject <{subj}> max version: {subj_schema.version}')

            yield subj_schema

    @staticmethod
    def _create_table(
        schema: Dict[str, Any],
        subject_name: str,
        cluster_name: str,
        schema_name: str,
        schema_description: str,
    ) -> Optional[TableMetadata]:
        """
        Create TableMetadata based on given schema and names
        """
        columns: List[ColumnMetadata] = []

        for i, field in enumerate(schema['fields']):
            columns.append(
                ColumnMetadata(
                    name=field['name'],
                    description=field.get('doc', None),
                    col_type=KafkaSchemaRegistryExtractor._get_property_type(
                        field
                    ),
                    sort_order=i,
                )
            )

        return TableMetadata(
            database='kafka_schema_registry',
            cluster=cluster_name,
            schema=subject_name,
            name=schema_name,
            description=schema_description,
            columns=columns,
        )

    @staticmethod
    def _get_property_type(schema: Dict) -> str:
        """
        Return type of the given schema.
        It will also works for nested schema types.
        """
        if 'type' not in schema:
            return 'object'

        if type(schema['type']) is dict:
            return KafkaSchemaRegistryExtractor._get_property_type(
                schema['type']
            )

        # If schema can have multiple types
        if type(schema['type']) is list:
            return '|'.join(schema['type'])

        if schema['type'] == 'record':
            properties = [
                f"{field['name']}:"
                f"{KafkaSchemaRegistryExtractor._get_property_type(field)}"
                for field in schema.get('fields', {})
            ]
            if len(properties) > 0:
                if 'name' in schema:
                    return schema['name'] + \
                        ':struct<' + ','.join(properties) + '>'
                return 'struct<' + ','.join(properties) + '>'
            return 'struct<object>'
        elif schema['type'] == 'array':
            items = KafkaSchemaRegistryExtractor._get_property_type(
                schema.get("items", {})
            )
            return 'array<' + items + '>'
        else:
            return schema['type']



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Iterator, Optional, Union,
)

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable

DESCRIPTION_NODE_LABEL_VAL = 'Description'
DESCRIPTION_NODE_LABEL = DESCRIPTION_NODE_LABEL_VAL


class DescriptionMetadata(GraphSerializable, AtlasSerializable):
    DESCRIPTION_NODE_LABEL = DESCRIPTION_NODE_LABEL_VAL
    PROGRAMMATIC_DESCRIPTION_NODE_LABEL = 'Programmatic_Description'
    DESCRIPTION_KEY_FORMAT = '{description}'
    DESCRIPTION_TEXT = 'description'
    DESCRIPTION_SOURCE = 'description_source'

    DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
    INVERSE_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION_OF'

    # The default editable source.
    DEFAULT_SOURCE = "description"

    def __init__(self,
                 text: Optional[str],
                 source: str = DEFAULT_SOURCE,
                 description_key: Optional[str] = None,
                 start_label: Optional[str] = None,  # Table, Column, Schema, Type_Metadata
                 start_key: Optional[str] = None,
                 ):
        """
        :param source: The unique source of what is populating this description.
        :param text: the description text. Markdown supported.
        """
        self.source = source
        self.text = text
        #  There are so many dependencies on Description node, that it is probably easier to just separate the rest out.
        if self.source == self.DEFAULT_SOURCE:
            self.label = self.DESCRIPTION_NODE_LABEL
        else:
            self.label = self.PROGRAMMATIC_DESCRIPTION_NODE_LABEL

        self.start_label = start_label
        self.start_key = start_key
        self.description_key = description_key or self.get_description_default_key(start_key)

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DescriptionMetadata):
            return (self.text == other.text and
                    self.source == other.source and
                    self.description_key == other.description_key and
                    self.start_label == other.start_label and
                    self.start_key == self.start_key)
        return False

    @staticmethod
    def create_description_metadata(text: Union[None, str],
                                    source: Optional[str] = DEFAULT_SOURCE,
                                    description_key: Optional[str] = None,
                                    start_label: Optional[str] = None,  # Table, Column, Schema
                                    start_key: Optional[str] = None,
                                    ) -> Optional['DescriptionMetadata']:
        # We do not want to create a node if there is no description text!
        if text is None:
            return None
        description_node = DescriptionMetadata(text=text,
                                               source=source or DescriptionMetadata.DEFAULT_SOURCE,
                                               description_key=description_key,
                                               start_label=start_label,
                                               start_key=start_key)
        return description_node

    def get_description_id(self) -> str:
        if self.source == self.DEFAULT_SOURCE:
            return "_description"
        else:
            return "_" + self.source + "_description"

    def get_description_default_key(self, start_key: Optional[str]) -> Optional[str]:
        return f'{start_key}/{self.get_description_id()}' if start_key else None

    def get_node(self, node_key: str) -> GraphNode:
        node = GraphNode(
            key=node_key,
            label=self.label,
            attributes={
                DescriptionMetadata.DESCRIPTION_SOURCE: self.source,
                DescriptionMetadata.DESCRIPTION_TEXT: self.text
            }
        )
        return node

    def get_relation(self,
                     start_node: str,
                     start_key: str,
                     end_key: str,
                     ) -> GraphRelationship:
        relationship = GraphRelationship(
            start_label=start_node,
            start_key=start_key,
            end_label=self.label,
            end_key=end_key,
            type=DescriptionMetadata.DESCRIPTION_RELATION_TYPE,
            reverse_type=DescriptionMetadata.INVERSE_DESCRIPTION_RELATION_TYPE,
            attributes={}
        )
        return relationship

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        if not self.description_key:
            raise Exception('Required description node key cannot be None')
        yield self.get_node(self.description_key)

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        if not self.start_label:
            raise Exception('Required relation start node label cannot be None')
        if not self.start_key:
            raise Exception('Required relation start key cannot be None')
        if not self.description_key:
            raise Exception('Required relation end key cannot be None')
        yield self.get_relation(
            start_node=self.start_label,
            start_key=self.start_key,
            end_key=self.description_key
        )

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        pass

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        pass

    def __repr__(self) -> str:
        return f'DescriptionMetadata({self.source!r}, {self.text!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Dict, Iterator, List, Union,
)

from pyhocon import ConfigTree
from simple_salesforce import Salesforce

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

LOGGER = logging.getLogger(__name__)


class SalesForceExtractor(Extractor):
    """
    Extracts SalesForce objects
    """

    # CONFIG KEYS
    CLUSTER_KEY = 'cluster_key'
    SCHEMA_KEY = 'schema_key'
    DATABASE_KEY = 'database_key'
    OBJECT_NAMES_KEY = "object_names"
    USERNAME_KEY = "username"
    PASSWORD_KEY = "password"
    SECURITY_TOKEN_KEY = "security_token"

    def init(self, conf: ConfigTree) -> None:

        self._cluster: str = conf.get_string(SalesForceExtractor.CLUSTER_KEY, "gold")
        self._database: str = conf.get_string(SalesForceExtractor.DATABASE_KEY)
        self._schema: str = conf.get_string(SalesForceExtractor.SCHEMA_KEY)
        self._object_names: List[str] = conf.get_list(SalesForceExtractor.OBJECT_NAMES_KEY, [])

        self._client: Salesforce = Salesforce(
            username=conf.get_string(SalesForceExtractor.USERNAME_KEY),
            password=conf.get_string(SalesForceExtractor.PASSWORD_KEY),
            security_token=conf.get_string(SalesForceExtractor.SECURITY_TOKEN_KEY),
        )

        self._extract_iter: Union[None, Iterator] = None

    def extract(self) -> Union[TableMetadata, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _get_extract_iter(self) -> Iterator[TableMetadata]:
        """
        Extract the TableMetaData for each SalesForce Object
        :return:
        """

        # Filter the sobjects if `OBJECT_NAMES_KEY` is set otherwise return all
        sobjects = [
            sobject
            for sobject in self._client.describe()["sobjects"]
            if (len(self._object_names) == 0 or sobject["name"] in self._object_names)
        ]

        for i, sobject in enumerate(sobjects):
            object_name = sobject["name"]
            logging.info(
                f"({i+1}/{len(sobjects)}) Extracting SalesForce object ({object_name})"
            )
            data = self._client.restful(path=f"sobjects/{object_name}/describe")
            yield self._extract_table_metadata(object_name=object_name, data=data)

    def _extract_table_metadata(
        self, object_name: str, data: Dict[str, Any]
    ) -> TableMetadata:
        # sort the fields by name because Amundsen requires a sort order for the columns and I did
        # not see one in the response
        fields = sorted(data["fields"], key=lambda x: x["name"])
        columns = [
            ColumnMetadata(
                name=f["name"],
                description=f["inlineHelpText"],
                col_type=f["type"],
                sort_order=i,
            )
            for i, f in enumerate(fields)
        ]
        return TableMetadata(
            database=self._database,
            cluster=self._cluster,
            schema=self._schema,
            name=object_name,
            # TODO: Can we extract table description / does it exist?
            description=None,
            columns=columns,
        )

    def get_scope(self) -> str:
        return 'extractor.salesforce_metadata'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Any

from pyhocon import ConfigTree

from databuilder import Scoped


class Extractor(Scoped):
    """
    An extractor extracts record
    """

    @abc.abstractmethod
    def init(self, conf: ConfigTree) -> None:
        pass

    @abc.abstractmethod
    def extract(self) -> Any:
        """
        :return: Provides a record or None if no more to extract
        """
        return None

    def get_scope(self) -> str:
        return 'extractor'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

API_VERSION = 'api_version'
API_BASE_URL = 'api_base_url'
TABLEAU_BASE_URL = 'tableau_base_url'
SITE_NAME = 'site_name'
TABLEAU_ACCESS_TOKEN_NAME = 'tableau_personal_access_token_name'
TABLEAU_ACCESS_TOKEN_SECRET = 'tableau_personal_access_token_secret'
EXCLUDED_PROJECTS = 'excluded_projects'
EXTERNAL_CLUSTER_NAME = 'external_cluster_name'
EXTERNAL_SCHEMA_NAME = 'external_schema_name'
EXTERNAL_TABLE_TYPES = 'external_table_types'
CLUSTER = 'cluster'
DATABASE = 'database'
VERIFY_REQUEST = 'verify_request'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator

from databuilder.models.graph_serializable import GraphSerializable


class QueryBase(GraphSerializable):
    @staticmethod
    def _normalize(sql: str) -> str:
        """
        Normalizes a SQL query or SQL expression.

        No checks are made to ensure that the input is valid SQL.
        This is not a full normalization.  The following operations are preformed:

        - Any run of whitespace characters outside of a quoted region is replaces by a single ' ' character.
        - Characters outside of quoted regions are made lower case.
        - If present, a trailing ';' is removed from the query.

        Note:
        Making characters outside quoted regions lower case does not in general result in an equivalent SQL statement.
        For example, with MySQL the case sensitivity of table names is operating system dependant.
        In practice, modern systems rarely rely on case sensitivity, and since making the non-quoted regions of the
        query lowercase is very helpful in identifying queries, we go ahead and do so.

        Also, this method fails to identify expressions such as `1 + 2` and `1+2`.
        There are likely too many special cases in this area to make much progress without doing a proper parse.
        """
        text = sql.strip()
        it = iter(text)
        sb = []
        for c in it:
            if c.isspace():
                c = QueryBase._process_whitespace(it)
                sb.append(' ')
            sb.append(c.lower())
            if c in ('`', '"', "'"):
                for d in QueryBase._process_quoted(it, c):
                    sb.append(d)
        if sb[-1] == ';':
            sb.pop()
        return ''.join(sb)

    @staticmethod
    def _process_quoted(it: Iterator[str], quote: str) -> Iterator[str]:
        """
        Yields characters up to and including the first occurrence of the (non-escaped) character `quote`.

        Allows `quote` to be escaped with '\\'.
        """
        p = ''
        for c in it:
            yield c
            if c == quote and p != '\\':
                break
            p = c

    @staticmethod
    def _process_whitespace(it: Iterator[str]) -> str:
        """
        Returns the first non-whitespace character encountered.

        This should never return `None` since the query text is striped before being processed.
        That is, if the current character is a whitespace character, then there remains at least one non-whitespace
        character in the stream.
        """
        for c in it:
            if not c.isspace():
                return c
        raise ValueError("Input string was not stripped!")



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_utils import ModeDashboardUtils
from databuilder.extractor.restapi.rest_api_extractor import MODEL_CLASS
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery

LOGGER = logging.getLogger(__name__)


class ModeDashboardOwnerExtractor(Extractor):
    """
    An Extractor that extracts Dashboard owner.

    """

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf

        restapi_query = self._build_restapi_query()
        self._extractor = ModeDashboardUtils.create_mode_rest_api_extractor(
            restapi_query=restapi_query,
            conf=self._conf.with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_owner.DashboardOwner', }
                )
            )
        )

    def extract(self) -> Any:
        return self._extractor.extract()

    def get_scope(self) -> str:
        return 'extractor.mode_dashboard_owner'

    def _build_restapi_query(self) -> ModePaginatedRestApiQuery:
        """
        Build REST API Query to get Mode Dashboard owner
        :return: A RestApiQuery that provides Mode Dashboard owner
        """

        seed_query = ModeDashboardUtils.get_seed_query(conf=self._conf)
        params = ModeDashboardUtils.get_auth_params(conf=self._conf, discover_auth=True)

        # Reports
        # https://mode.com/developer/discovery-api/analytics/reports/
        url = 'https://app.mode.com/batch/{organization}/reports'
        json_path = 'reports[*].[token, space_token, creator_email]'
        field_names = ['dashboard_id', 'dashboard_group_id', 'email']
        max_record_size = 1000
        pagination_json_path = 'reports[*]'
        creator_query = ModePaginatedRestApiQuery(query_to_join=seed_query, url=url, params=params,
                                                  json_path=json_path, field_names=field_names,
                                                  skip_no_result=True, max_record_size=max_record_size,
                                                  pagination_json_path=pagination_json_path)

        return creator_query



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

class PublisherConfigs:
    # A directory that contains CSV files for nodes
    NODE_FILES_DIR = 'node_files_directory'
    # A directory that contains CSV files for relationships
    RELATION_FILES_DIR = 'relation_files_directory'

    # A CSV header with this suffix will be passed to the statement without quotes
    UNQUOTED_SUFFIX = ':UNQUOTED'

    # This will be used to provide unique tag to the node and relationship
    JOB_PUBLISH_TAG = 'job_publish_tag'

    # any additional fields that should be added to nodes and rels through config
    ADDITIONAL_PUBLISHER_METADATA_FIELDS = 'additional_publisher_metadata_fields'

    # Property name for published tag
    PUBLISHED_TAG_PROPERTY_NAME = 'published_tag'
    # Property name for last updated timestamp
    LAST_UPDATED_EPOCH_MS = 'publisher_last_updated_epoch_ms'


class PublishBehaviorConfigs:
    # A boolean flag to indicate if publisher_metadata (e.g. published_tag,
    # publisher_last_updated_epoch_ms)
    # will be included as properties of the nodes
    ADD_PUBLISHER_METADATA = 'add_publisher_metadata'

    # NOTE: Do not use this unless you have a specific use case for it. Amundsen expects two way relationships, and
    # the default value should be set to true to publish relations in both directions. If it is overridden and set
    # to false, reverse relationships will not be published.
    PUBLISH_REVERSE_RELATIONSHIPS = 'publish_reverse_relationships'

    # If enabled, stops the publisher from updating a node or relationship
    # created via the UI, e.g. a description or owner added manually by an Amundsen user.
    # Such nodes/relationships will not have a 'published_tag' property that is set by databuilder.
    PRESERVE_ADHOC_UI_DATA = 'preserve_adhoc_ui_data'

    # If enabled, the default behavior will continue to publish properties with empty values.
    # If False, empty properties will be set to NULL and will not show up on the node or relation.
    PRESERVE_EMPTY_PROPS = 'preserve_empty_props'


class Neo4jCsvPublisherConfigs:
    # A end point for Neo4j e.g: bolt://localhost:9999
    NEO4J_END_POINT_KEY = 'neo4j_endpoint'
    # A transaction size that determines how often it commits.
    NEO4J_TRANSACTION_SIZE = 'neo4j_transaction_size'

    NEO4J_MAX_CONN_LIFE_TIME_SEC = 'neo4j_max_conn_life_time_sec'

    # list of nodes that are create only, and not updated if match exists
    NEO4J_CREATE_ONLY_NODES = 'neo4j_create_only_nodes'

    NEO4J_USER = 'neo4j_user'
    NEO4J_PASSWORD = 'neo4j_password'
    # in Neo4j (v4.0+), we can create and use more than one active database at the same time
    NEO4J_DATABASE_NAME = 'neo4j_database'

    # NEO4J_ENCRYPTED is a boolean indicating whether to use SSL/TLS when connecting
    NEO4J_ENCRYPTED = 'neo4j_encrypted'
    # NEO4J_VALIDATE_SSL is a boolean indicating whether to validate the server's SSL/TLS
    # cert against system CAs
    NEO4J_VALIDATE_SSL = 'neo4j_validate_ssl'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


from typing import (
    Any, Dict, Iterator, List, Set, Union,
)

from databuilder.extractor.es_base_extractor import ElasticsearchBaseExtractor
from databuilder.models.table_stats import TableColumnStats


class ElasticsearchColumnStatsExtractor(ElasticsearchBaseExtractor):
    """
    Extractor to extract stats for Elasticsearch mapping attributes.
    """

    def get_scope(self) -> str:
        return 'extractor.es_column_stats'

    def _get_index_stats(self, index_name: str, fields: List[str]) -> List[Dict[str, Any]]:
        query = {
            "size": 0,
            "aggs": {
                "stats": {
                    "matrix_stats": {
                        "fields": fields
                    }
                }
            }
        }

        _data = self.es.search(index=index_name, body=query)

        data = _data.get('aggregations', dict()).get('stats', dict()).get('fields', list())

        return data

    def _render_column_stats(self, index_name: str, spec: Dict[str, Any]) -> List[TableColumnStats]:
        result: List[TableColumnStats] = []

        col_name = spec.pop('name')

        for stat_name, stat_val in spec.items():
            if isinstance(stat_val, dict) or isinstance(stat_val, list):
                continue
            elif stat_val == 'NaN':
                continue

            stat = TableColumnStats(table_name=index_name,
                                    col_name=col_name,
                                    stat_name=stat_name,
                                    stat_val=stat_val,
                                    start_epoch='0',
                                    end_epoch='0',
                                    db=self.database,
                                    cluster=self.cluster,
                                    schema=self.schema)

            result.append(stat)

        return result

    @property
    def _allowed_types(self) -> Set[str]:
        return set(['long', 'integer', 'short', 'byte', 'double',
                    'float', 'half_float', 'scaled_float', 'unsigned_long'])

    def _get_extract_iter(self) -> Iterator[Union[TableColumnStats, None]]:
        indexes: Dict = self._get_indexes()
        for index_name, index_metadata in indexes.items():
            properties = self._get_index_mapping_properties(index_metadata) or dict()

            fields = [name for name, spec in properties.items() if spec['type'] in self._allowed_types]

            specifications = self._get_index_stats(index_name, fields)

            for spec in specifications:
                stats = self._render_column_stats(index_name, spec)

                for stat in stats:
                    yield stat



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import abc
from typing import (
    Any, Dict, Iterator, List, Optional, Union,
)

from pyhocon import ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata


class ElasticsearchBaseExtractor(Extractor):
    """
    Extractor to extract index metadata from Elasticsearch

    By default, the extractor does not add sort_order to columns. Set ELASTICSEARCH_CORRECT_SORT_ORDER conf to True
    for columns to have correct sort order.

    Set ELASTICSEARCH_TIME_FIELD to the name of the field representing time.
    """

    ELASTICSEARCH_CLIENT_CONFIG_KEY = 'client'
    ELASTICSEARCH_EXTRACT_TECHNICAL_DETAILS = 'extract_technical_details'

    # For backwards compatibility, the Elasticsearch extractor does not add sort_order to columns by default.
    # Set this to true in the conf for columns to have correct sort order.
    ELASTICSEARCH_CORRECT_SORT_ORDER = 'correct_sort_order'

    # Set this to the name of the field representing time.
    ELASTICSEARCH_TIME_FIELD = 'time_field'

    CLUSTER = 'cluster'
    SCHEMA = 'schema'

    def __init__(self) -> None:
        super(ElasticsearchBaseExtractor, self).__init__()

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf
        self._extract_iter = self._get_extract_iter()

        self.es = self.conf.get(ElasticsearchBaseExtractor.ELASTICSEARCH_CLIENT_CONFIG_KEY)

    def _get_es_version(self) -> str:
        return self.es.info().get('version').get('number')

    def _get_indexes(self) -> Dict:
        result = dict()

        try:
            _indexes = self.es.indices.get('*')

            for k, v in _indexes.items():
                if not k.startswith('.'):
                    result[k] = v
        except Exception:
            pass

        return result

    def _get_index_creation_date(self, index_metadata: Dict) -> Optional[float]:
        try:
            return float(index_metadata.get('settings', dict()).get('index').get('creation_date'))
        except Exception:
            return None

    def _get_index_mapping_properties(self, index: Dict) -> Optional[Dict]:
        mappings = index.get('mappings', dict())

        # Mapping types were removed in Elasticsearch 7. As a result, index mappings are formatted differently.
        # See https://www.elastic.co/guide/en/elasticsearch/reference/current/removal-of-types.html
        version = self._get_es_version()

        try:
            if int(version.split('.')[0]) >= 7:
                properties = mappings.get('properties', dict())
            else:
                properties = list(mappings.values())[0].get('properties', dict())
        except Exception:
            properties = dict()

        return properties

    def _get_attributes(self,
                        input_mapping: Dict,
                        parent_col_name: str = '',
                        separator: str = '.') -> List[ColumnMetadata]:
        cols: List[ColumnMetadata] = []

        for col_name, col_mapping in input_mapping.items():
            qualified_col_name = str(parent_col_name) + separator + col_name if parent_col_name else col_name
            if isinstance(col_mapping, dict):
                if col_mapping.__contains__('properties'):
                    # Need to recurse
                    inner_mapping: Dict = col_mapping.get('properties', {})
                    cols.extend(self._get_attributes(input_mapping=inner_mapping,
                                                     parent_col_name=qualified_col_name,
                                                     separator=separator))
                else:
                    cols.append(ColumnMetadata(name=qualified_col_name,
                                               description='',
                                               col_type=col_mapping.get('type', ''),
                                               sort_order=0))

        return cols

    def extract(self) -> Any:
        try:
            result = next(self._extract_iter)

            return result
        except StopIteration:
            return None

    @property
    def database(self) -> str:
        return 'elasticsearch'

    @property
    def cluster(self) -> str:
        return self.conf.get(ElasticsearchBaseExtractor.CLUSTER)

    @property
    def schema(self) -> str:
        return self.conf.get(ElasticsearchBaseExtractor.SCHEMA)

    @property
    def _extract_technical_details(self) -> bool:
        try:
            return self.conf.get(ElasticsearchBaseExtractor.ELASTICSEARCH_EXTRACT_TECHNICAL_DETAILS)
        except Exception:
            return False

    @property
    def _correct_sort_order(self) -> bool:
        try:
            return self.conf.get(ElasticsearchBaseExtractor.ELASTICSEARCH_CORRECT_SORT_ORDER)
        except Exception:
            return False

    # Default time field is @timestamp to match ECS
    # See https://www.elastic.co/guide/en/ecs/master/ecs-base.html
    @property
    def _time_field(self) -> str:
        return self.conf.get(ElasticsearchBaseExtractor.ELASTICSEARCH_TIME_FIELD, '@timestamp')

    @abc.abstractmethod
    def _get_extract_iter(self) -> Iterator[Union[Any, None]]:
        pass



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
import logging
import multiprocessing.pool
from copy import deepcopy
from functools import reduce
from typing import (
    Any, Dict, Generator, Iterator, List, Optional, Tuple,
)

from amundsen_common.utils.atlas import AtlasTableKey
from atlasclient.client import Atlas
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor

LOGGER = logging.getLogger(__name__)

# custom types
type_fields_mapping_spec = Dict[str, List[Tuple[str, Any, Any, Any]]]
type_fields_mapping = List[Tuple[str, Any, Any, Any]]

# @todo document classes/methods
# @todo write tests

__all__ = ['AtlasSearchDataExtractor']


class AtlasSearchDataExtractorHelpers:
    @staticmethod
    def _filter_none(input_list: List) -> List:
        return list(filter(None, input_list))

    @staticmethod
    def get_entity_names(entity_list: Optional[List]) -> List:
        entity_list = entity_list or []
        return AtlasSearchDataExtractorHelpers._filter_none(
            [e.get('attributes').get('name') for e in entity_list if e.get('status').lower() == 'active'])

    @staticmethod
    def get_entity_uri(qualified_name: str, type_name: str) -> str:
        key = AtlasTableKey(qualified_name, database=type_name)
        return key.amundsen_key

    @staticmethod
    def get_entity_descriptions(entity_list: Optional[List]) -> List:
        entity_list = entity_list or []
        return AtlasSearchDataExtractorHelpers._filter_none(
            [e.get('attributes', dict()).get('description') for e in entity_list
             if e.get('status').lower() == 'active'])

    @staticmethod
    def get_badges_from_classifications(classifications: Optional[List]) -> List:
        classifications = classifications or []
        return AtlasSearchDataExtractorHelpers._filter_none(
            [c.get('typeName') for c in classifications if c.get('entityStatus', '').lower() == 'active'])

    @staticmethod
    def get_display_text(meanings: Optional[List]) -> List:
        meanings = meanings or []
        return AtlasSearchDataExtractorHelpers._filter_none(
            [c.get('displayText') for c in meanings if c.get('entityStatus', '').lower() == 'active'])

    @staticmethod
    def get_last_successful_execution_timestamp(executions: Optional[List]) -> int:
        executions = executions or []
        successful_executions = AtlasSearchDataExtractorHelpers._filter_none(
            [e.get('attributes').get('timestamp') for e in executions
             if e.get('status', '').lower() == 'active' and e.get('attributes', dict()).get('state') == 'succeeded'])

        try:
            return max(successful_executions)
        except ValueError:
            return 0

    @staticmethod
    def get_chart_names(queries: Optional[List]) -> List[str]:
        queries = queries or []
        charts = []

        for query in queries:
            _charts = query.get('relationshipAttributes', dict()).get('charts', [])
            charts += _charts

        return AtlasSearchDataExtractorHelpers.get_display_text(charts)

    @staticmethod
    def get_table_database(table_key: str) -> str:
        result = AtlasTableKey(table_key).get_details().get('database', 'hive_table')

        return result

    @staticmethod
    def get_source_description(parameters: Optional[dict]) -> str:
        parameters = parameters or dict()

        return parameters.get('sourceDescription', '')

    @staticmethod
    def get_usage(readers: Optional[List]) -> Tuple[int, int]:
        readers = readers or []

        score = 0
        unique = 0

        for reader in readers:
            reader_status = reader.get('status')
            entity_status = reader.get('relationshipAttributes', dict()).get('entity', dict()).get('entityStatus', '')
            relationship_status = reader.get('relationshipAttributes',
                                             dict()).get('entity',
                                                         dict()).get('relationshipStatus', '')

            if reader_status == entity_status == relationship_status == 'ACTIVE':
                score += reader.get('attributes', dict()).get('count', 0)

                if score > 0:
                    unique += 1

        return score, unique


class AtlasSearchDataExtractor(Extractor):
    ATLAS_URL_CONFIG_KEY = 'atlas_url'
    ATLAS_PORT_CONFIG_KEY = 'atlas_port'
    ATLAS_PROTOCOL_CONFIG_KEY = 'atlas_protocol'
    ATLAS_VALIDATE_SSL_CONFIG_KEY = 'atlas_validate_ssl'
    ATLAS_USERNAME_CONFIG_KEY = 'atlas_auth_user'
    ATLAS_PASSWORD_CONFIG_KEY = 'atlas_auth_pw'
    ATLAS_SEARCH_CHUNK_SIZE_KEY = 'atlas_search_chunk_size'
    ATLAS_DETAILS_CHUNK_SIZE_KEY = 'atlas_details_chunk_size'
    ATLAS_TIMEOUT_SECONDS_KEY = 'atlas_timeout_seconds'
    ATLAS_MAX_RETRIES_KEY = 'atlas_max_retries'

    PROCESS_POOL_SIZE_KEY = 'process_pool_size'

    ENTITY_TYPE_KEY = 'entity_type'

    DEFAULT_CONFIG = ConfigFactory.from_dict({ATLAS_URL_CONFIG_KEY: "localhost",
                                              ATLAS_PORT_CONFIG_KEY: 21000,
                                              ATLAS_PROTOCOL_CONFIG_KEY: 'http',
                                              ATLAS_VALIDATE_SSL_CONFIG_KEY: False,
                                              ATLAS_SEARCH_CHUNK_SIZE_KEY: 250,
                                              ATLAS_DETAILS_CHUNK_SIZE_KEY: 25,
                                              ATLAS_TIMEOUT_SECONDS_KEY: 120,
                                              ATLAS_MAX_RETRIES_KEY: 2,
                                              PROCESS_POOL_SIZE_KEY: 10})

    # es_document field, atlas field path, modification function, default_value
    FIELDS_MAPPING_SPEC: type_fields_mapping_spec = {
        'Table': [
            ('database', 'attributes.qualifiedName',
             lambda x: AtlasSearchDataExtractorHelpers.get_table_database(x), None),
            ('cluster', 'attributes.qualifiedName',
             lambda x: AtlasTableKey(x).get_details()['cluster'], None),
            ('schema', 'attributes.qualifiedName',
             lambda x: AtlasTableKey(x).get_details()['schema'], None),
            ('name', 'attributes.name', None, None),
            ('key', ['attributes.qualifiedName', 'typeName'],
             lambda x, y: AtlasSearchDataExtractorHelpers.get_entity_uri(x, y), None),
            ('description', 'attributes.description', None, None),
            ('last_updated_timestamp', 'updateTime', lambda x: int(x) / 1000, 0),
            ('total_usage', 'relationshipAttributes.readers',
             lambda x: AtlasSearchDataExtractorHelpers.get_usage(x)[0], 0),
            ('unique_usage', 'relationshipAttributes.readers',
             lambda x: AtlasSearchDataExtractorHelpers.get_usage(x)[1], 0),
            ('column_names', 'relationshipAttributes.columns',
             lambda x: AtlasSearchDataExtractorHelpers.get_entity_names(x), []),
            ('column_descriptions', 'relationshipAttributes.columns',
             lambda x: AtlasSearchDataExtractorHelpers.get_entity_descriptions(x), []),
            ('tags', 'relationshipAttributes.meanings',
             lambda x: AtlasSearchDataExtractorHelpers.get_display_text(x), []),
            ('badges', 'classifications',
             lambda x: AtlasSearchDataExtractorHelpers.get_badges_from_classifications(x), []),
            ('display_name', 'attributes.qualifiedName',
             lambda x: '.'.join([AtlasTableKey(x).get_details()['schema'], AtlasTableKey(x).get_details()['table']]),
             None),
            ('schema_description', 'attributes.parameters',
             lambda x: AtlasSearchDataExtractorHelpers.get_source_description(x), ''),
            ('programmatic_descriptions', 'attributes.parameters', lambda x: [str(s) for s in list(x.values())], {})
        ],
        'Dashboard': [
            ('group_name', 'relationshipAttributes.group.attributes.name', None, None),
            ('name', 'attributes.name', None, None),
            ('description', 'attributes.description', None, None),
            ('total_usage', 'relationshipAttributes.readers',
             lambda x: AtlasSearchDataExtractorHelpers.get_usage(x)[0], 0),
            ('product', 'attributes.product', None, None),
            ('cluster', 'attributes.cluster', None, None),
            ('group_description', 'relationshipAttributes.group.attributes.description', None, None),
            ('query_names', 'relationshipAttributes.queries',
             lambda x: AtlasSearchDataExtractorHelpers.get_entity_names(x), []),
            ('chart_names', 'relationshipAttributes.queries',
             lambda x: AtlasSearchDataExtractorHelpers.get_chart_names(x), []),
            ('group_url', 'relationshipAttributes.group.attributes.url', None, None),
            ('url', 'attributes.url', None, None),
            ('uri', 'attributes.qualifiedName', None, None),
            ('last_successful_run_timestamp', 'relationshipAttributes.executions',
             lambda x: AtlasSearchDataExtractorHelpers.get_last_successful_execution_timestamp(x), None),
            ('tags', 'relationshipAttributes.meanings',
             lambda x: AtlasSearchDataExtractorHelpers.get_display_text(x), []),
            ('badges', 'classifications',
             lambda x: AtlasSearchDataExtractorHelpers.get_badges_from_classifications(x), [])
        ],
        'User': [
            ('email', 'attributes.qualifiedName', None, ''),
            ('first_name', 'attributes.first_name', None, ''),
            ('last_name', 'attributes.last_name', None, ''),
            ('full_name', 'attributes.full_name', None, ''),
            ('github_username', 'attributes.github_username', None, ''),
            ('team_name', 'attributes.team_name', None, ''),
            ('employee_type', 'attributes.employee_type', None, ''),
            ('manager_email', 'attributes.manager_email', None, ''),
            ('slack_id', 'attributes.slack_id', None, ''),
            ('role_name', 'attributes.role_name', None, ''),
            ('is_active', 'attributes.is_active', None, ''),
            ('total_read', 'attributes.total_read', None, ''),
            ('total_own', 'attributes.total_own', None, ''),
            ('total_follow', 'attributes.total_follow', None, '')
        ]
    }

    ENTITY_MODEL_BY_TYPE = {
        'Table': 'databuilder.models.table_elasticsearch_document.TableESDocument',
        'Dashboard': 'databuilder.models.dashboard_elasticsearch_document.DashboardESDocument',
        'User': 'databuilder.models.user_elasticsearch_document.UserESDocument'
    }

    REQUIRED_RELATIONSHIPS_BY_TYPE = {
        'Table': ['columns', 'readers'],
        'Dashboard': ['group', 'charts', 'executions', 'queries'],
        'User': []
    }

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf.with_fallback(AtlasSearchDataExtractor.DEFAULT_CONFIG)
        self.driver = self._get_driver()

        self._extract_iter: Optional[Iterator[Any]] = None

    @property
    def entity_type(self) -> str:
        return self.conf.get(AtlasSearchDataExtractor.ENTITY_TYPE_KEY)

    @property
    def dsl_search_query(self) -> Dict:
        query = {
            'query': f'{self.entity_type} where __state = "ACTIVE"'
        }

        LOGGER.debug(f'DSL Search Query: {query}')

        return query

    @property
    def model_class(self) -> Any:
        model_class = AtlasSearchDataExtractor.ENTITY_MODEL_BY_TYPE.get(self.entity_type)

        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)

            return getattr(mod, class_name)

    @property
    def field_mappings(self) -> type_fields_mapping:
        return AtlasSearchDataExtractor.FIELDS_MAPPING_SPEC.get(self.entity_type) or []

    @property
    def search_chunk_size(self) -> int:
        return self.conf.get_int(AtlasSearchDataExtractor.ATLAS_SEARCH_CHUNK_SIZE_KEY)

    @property
    def relationships(self) -> Optional[List[str]]:
        return AtlasSearchDataExtractor.REQUIRED_RELATIONSHIPS_BY_TYPE.get(self.entity_type)  # type: ignore

    def extract(self) -> Any:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()

        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.atlas_search_data'

    def _get_driver(self) -> Any:
        return Atlas(host=self.conf.get_string(AtlasSearchDataExtractor.ATLAS_URL_CONFIG_KEY),
                     port=self.conf.get_string(AtlasSearchDataExtractor.ATLAS_PORT_CONFIG_KEY),
                     username=self.conf.get_string(AtlasSearchDataExtractor.ATLAS_USERNAME_CONFIG_KEY),
                     password=self.conf.get_string(AtlasSearchDataExtractor.ATLAS_PASSWORD_CONFIG_KEY),
                     protocol=self.conf.get_string(AtlasSearchDataExtractor.ATLAS_PROTOCOL_CONFIG_KEY),
                     validate_ssl=self.conf.get_bool(AtlasSearchDataExtractor.ATLAS_VALIDATE_SSL_CONFIG_KEY),
                     timeout=self.conf.get_int(AtlasSearchDataExtractor.ATLAS_TIMEOUT_SECONDS_KEY),
                     max_retries=self.conf.get_int(AtlasSearchDataExtractor.ATLAS_MAX_RETRIES_KEY))

    def _get_latest_entity_metrics(self) -> Optional[dict]:
        admin_metrics = list(self.driver.admin_metrics)

        try:
            return admin_metrics[-1].entity
        except Exception:
            return None

    def _get_count_of_active_entities(self) -> int:
        entity_metrics = self._get_latest_entity_metrics()

        if entity_metrics:
            count = entity_metrics.get('entityActive-typeAndSubTypes', dict()).get(self.entity_type, 0)

            return int(count)
        else:
            return 0

    def _get_entity_guids(self, start_offset: int) -> List[str]:
        result = []

        batch_start = start_offset
        batch_end = start_offset + self.search_chunk_size

        LOGGER.info(f'Collecting guids for batch: {batch_start}-{batch_end}')

        _params = {'offset': str(batch_start), 'limit': str(self.search_chunk_size)}

        full_params = deepcopy(self.dsl_search_query)
        full_params.update(**_params)

        try:
            results = self.driver.search_dsl(**full_params)

            for hit in results:
                for entity in hit.entities:
                    result.append(entity.guid)

            return result
        except Exception:
            LOGGER.warning(f'Error processing batch: {batch_start}-{batch_end}', exc_info=True)

            return []

    def _get_entity_details(self, guid_list: List[str]) -> List:
        result = []

        LOGGER.info(f'Processing guids chunk of size: {len(guid_list)}')

        try:
            bulk_collection = self.driver.entity_bulk(guid=guid_list)

            for collection in bulk_collection:
                search_chunk = list(collection.entities_with_relationships(attributes=self.relationships))

                result += search_chunk

            return result
        except Exception:
            LOGGER.warning(f'Error processing guids. {len(guid_list)}', exc_info=True)

            return []

    @staticmethod
    def split_list_to_chunks(input_list: List[Any], n: int) -> Generator:
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(input_list), n):
            yield input_list[i:i + n]

    def _execute_query(self) -> Any:
        details_chunk_size = self.conf.get_int(AtlasSearchDataExtractor.ATLAS_DETAILS_CHUNK_SIZE_KEY)
        process_pool_size = self.conf.get_int(AtlasSearchDataExtractor.PROCESS_POOL_SIZE_KEY)

        guids = []

        entity_count = self._get_count_of_active_entities()

        LOGGER.info(f'Received count: {entity_count}')

        if entity_count > 0:
            offsets = [i * self.search_chunk_size for i in range(int(entity_count / self.search_chunk_size) + 1)]
        else:
            offsets = []

        with multiprocessing.pool.ThreadPool(processes=process_pool_size) as pool:
            guid_list = pool.map(self._get_entity_guids, offsets, chunksize=1)

        for sub_list in guid_list:
            guids += sub_list

        LOGGER.info(f'Received guids: {len(guids)}')

        if guids:
            guids_chunks = AtlasSearchDataExtractor.split_list_to_chunks(guids, details_chunk_size)

            with multiprocessing.pool.ThreadPool(processes=process_pool_size) as pool:
                return_list = pool.map(self._get_entity_details, guids_chunks)

            for sub_list in return_list:
                for entry in sub_list:
                    yield entry

    def _get_extract_iter(self) -> Iterator[Any]:
        for atlas_entity in self._execute_query():
            model_dict = dict()

            try:
                data = atlas_entity.__dict__['_data']

                for spec in self.field_mappings:
                    model_field, atlas_fields_paths, _transform_spec, default_value = spec

                    if not isinstance(atlas_fields_paths, list):
                        atlas_fields_paths = [atlas_fields_paths]

                    atlas_values = []
                    for atlas_field_path in atlas_fields_paths:

                        atlas_value = reduce(lambda x, y: x.get(y, dict()), atlas_field_path.split('.'),
                                             data) or default_value
                        atlas_values.append(atlas_value)

                    transform_spec = _transform_spec or (lambda x: x)

                    es_entity_value = transform_spec(*atlas_values)
                    model_dict[model_field] = es_entity_value

                yield self.model_class(**model_dict)
            except Exception:
                LOGGER.warning('Error building model object.', exc_info=True)
